from typing import List

import pandas as pd
from mindsdb_sql_parser import ast

from mindsdb.integrations.utilities.handlers.query_utilities import (
    SELECTQueryExecutor,
    SELECTQueryParser,
)
from mindsdb.integrations.libs.api_handler import APIHandler, APITable
from mindsdb.integrations.utilities.sql_utils import conditions_to_filter


class CustomAPITable(APITable):

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.handler.connect()

    def get_columns(self, ignore: List[str] = []) -> List[str]:
        return [item for item in self.columns if item not in ignore]

    def select(self, query: ast.Select) -> pd.DataFrame:
        raise NotImplementedError()

    def parse_select(self, query: ast.Select, table_name: str):
        select_statement_parser = SELECTQueryParser(query, table_name, self.get_columns())
        self.selected_columns, self.where_conditions, self.order_by_conditions, self.result_limit = select_statement_parser.parse_query()

    def get_where_param(self, query: ast.Select, param: str):
        params = conditions_to_filter(query.where)
        if param not in params:
            raise Exception(f"WHERE condition does not have '{param}' selector")
        return params[param]

    def apply_query_params(self, df, query):
        select_statement_parser = SELECTQueryParser(query, self.name, self.get_columns())
        selected_columns, _, order_by_conditions, result_limit = select_statement_parser.parse_query()
        select_statement_executor = SELECTQueryExecutor(df, selected_columns, [], order_by_conditions, result_limit)
        return select_statement_executor.execute_query()


class AddressEmailAddressTable(CustomAPITable):
    """Email address data linked to all business partner address records in the system"""

    name: str = "address_email_address"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "AddressID",
        "Person",
        "OrdinalNumber",
        "IsDefaultEmailAddress",
        "EmailAddress",
        "SearchEmailAddress",
        "AddressCommunicationRemarkText",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_AddressEmailAddress")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class AddressFaxNumberTable(CustomAPITable):
    """Fax address data linked to all the business partner address records in the system"""

    name: str = "address_fax_number"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "AddressID",
        "Person",
        "OrdinalNumber",
        "IsDefaultFaxNumber",
        "FaxCountry",
        "FaxNumber",
        "FaxNumberExtension",
        "InternationalFaxNumber",
        "AddressCommunicationRemarkText",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_AddressFaxNumber")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class AddressHomePageURLTable(CustomAPITable):
    """Home page URL address records linked to all business partner address records in the system"""

    name: str = "address_home_page"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "AddressID",
        "Person",
        "OrdinalNumber",
        "ValidityStartDate",
        "IsDefaultURLAddress",
        "SearchURLAddress",
        "AddressCommunicationRemarkText",
        "URLFieldLength",
        "WebsiteURL",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_AddressHomePageURL")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class AddressPhoneNumberTable(CustomAPITable):
    """All the mobile/telephone address records linked to all the business partner address records in the system"""

    name: str = "address_phone_number"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "AddressID",
        "Person",
        "OrdinalNumber",
        "DestinationLocationCountry",
        "IsDefaultPhoneNumber",
        "PhoneNumber",
        "PhoneNumberExtension",
        "InternationalPhoneNumber",
        "PhoneNumberType",
        "AddressCommunicationRemarkText",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_AddressPhoneNumber")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class BPAddrDepdntIntlLocNumberTable(CustomAPITable):
    """address dependent data for the business partner address by using the key fields business partner number and address ID"""

    name: str = "bp_addr_depdnt_intl_loc_number"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "BusinessPartner",
        "AddressID",
        "InternationalLocationNumber1",
        "InternationalLocationNumber2",
        "InternationalLocationNumber3",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_BPAddrDepdntIntlLocNumber")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class BPContactToAddressTable(CustomAPITable):
    """Workplace address records linked to all the business partner contact records in the system."""

    name: str = "bp_contact_to_address"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "RelationshipNumber",
        "BusinessPartnerCompany",
        "BusinessPartnerPerson",
        "ValidityEndDate",
        "AddressID",
        "AddressNumber",
        "AdditionalStreetPrefixName",
        "AdditionalStreetSuffixName",
        "AddressTimeZone",
        "CareOfName",
        "CityCode",
        "CityName",
        "CompanyPostalCode",
        "Country",
        "County",
        "DeliveryServiceNumber",
        "DeliveryServiceTypeCode",
        "District",
        "FormOfAddress",
        "FullName",
        "HomeCityName",
        "HouseNumber",
        "HouseNumberSupplementText",
        "Language",
        "POBox",
        "POBoxDeviatingCityName",
        "POBoxDeviatingCountry",
        "POBoxDeviatingRegion",
        "POBoxIsWithoutNumber",
        "POBoxLobbyName",
        "POBoxPostalCode",
        "Person",
        "PostalCode",
        "PrfrdCommMediumType",
        "Region",
        "StreetName",
        "StreetPrefixName",
        "StreetSuffixName",
        "TaxJurisdiction",
        "TransportZone",
        "AddressRepresentationCode",
        "ContactPersonBuilding",
        "ContactPersonPrfrdCommMedium",
        "ContactRelationshipDepartment",
        "ContactRelationshipFunction",
        "CorrespondenceShortName",
        "Floor",
        "InhouseMail",
        "IsDefaultAddress",
        "RoomNumber",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_BPContactToAddress")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class BPContactToFuncAndDeptTable(CustomAPITable):
    """Contact person department and function data linked to all business partner contact records in the system"""

    name: str = "bp_contact_to_func_and_dept"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "RelationshipNumber",
        "BusinessPartnerCompany",
        "BusinessPartnerPerson",
        "ValidityEndDate",
        "ContactPersonAuthorityType",
        "ContactPersonDepartment",
        "ContactPersonDepartmentName",
        "ContactPersonFunction",
        "ContactPersonFunctionName",
        "ContactPersonRemarkText",
        "ContactPersonVIPType",
        "EmailAddress",
        "FaxNumber",
        "FaxNumberExtension",
        "PhoneNumber",
        "PhoneNumberExtension",
        "RelationshipCategory",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_BPContactToFuncAndDept")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class BPCreditWorthinessTable(CustomAPITable):
    """Contact person department and function data linked to all business partner contact records in the system"""

    name: str = "bp_credit_worthiness"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "BusinessPartner",
        "BusPartCreditStanding",
        "BPCreditStandingStatus",
        "CreditRatingAgency",
        "BPCreditStandingComment",
        "BPCreditStandingDate",
        "BPCreditStandingRating",
        "BPLegalProceedingStatus",
        "BPLglProceedingInitiationDate",
        "BusinessPartnerIsUnderOath",
        "BusinessPartnerOathDate",
        "BusinessPartnerIsBankrupt",
        "BusinessPartnerBankruptcyDate",
        "BPForeclosureIsInitiated",
        "BPForeclosureDate",
        "BPCrdtWrthnssAccessChkIsActive",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_BPCreditWorthiness")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class BPDataControllerTable(CustomAPITable):
    """Business partner data controllers of all the available records linked to business partners in the system"""

    name: str = "bp_data_controller"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "BusinessPartner",
        "DataController",
        "PurposeForPersonalData",
        "DataControlAssignmentStatus",
        "BPDataControllerIsDerived",
        "PurposeDerived",
        "PurposeType",
        "BusinessPurposeFlag",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_BPDataController")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class BPFinancialServicesExtnTable(CustomAPITable):
    """Financial services business partner attributes of all the available records linked to business partners in the system"""

    name: str = "bp_financial_services_extn"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "BusinessPartner",
        "BusinessPartnerIsVIP",
        "TradingPartner",
        "FactoryCalendar",
        "BusinessPartnerOfficeCountry",
        "BusinessPartnerOfficeRegion",
        "BPRegisteredOfficeName",
        "BPBalanceSheetCurrency",
        "BPLastCptlIncrAmtInBalShtCrcy",
        "BPLastCapitalIncreaseYear",
        "BPBalanceSheetDisplayType",
        "BusinessPartnerCitizenship",
        "BPMaritalPropertyRegime",
        "BusinessPartnerIncomeCurrency",
        "BPNumberOfChildren",
        "BPNumberOfHouseholdMembers",
        "BPAnnualNetIncAmtInIncomeCrcy",
        "BPMonthlyNetIncAmtInIncomeCrcy",
        "BPAnnualNetIncomeYear",
        "BPMonthlyNetIncomeMonth",
        "BPMonthlyNetIncomeYear",
        "BPPlaceOfDeathName",
        "CustomerIsUnwanted",
        "UndesirabilityReason",
        "UndesirabilityComment",
        "LastCustomerContactDate",
        "BPGroupingCharacter",
        "BPLetterSalutation",
        "BusinessPartnerTargetGroup",
        "BusinessPartnerEmployeeGroup",
        "BusinessPartnerIsEmployee",
        "BPTermnBusRelationsBankDate",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_BPFinancialServicesExtn")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class BPFinancialServicesReportingTable(CustomAPITable):
    """Financial services reporting attributes of all the available records linked to business partners in the system"""

    name: str = "bp_financial_services_reporting"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "BusinessPartner",
        "BPIsNonResident",
        "BPNonResidencyStartDate",
        "BPIsMultimillionLoanRecipient",
        "BPLoanReportingBorrowerNumber",
        "BPLoanRptgBorrowerEntityNumber",
        "BPCreditStandingReview",
        "BPCreditStandingReviewDate",
        "BusinessPartnerLoanToManager",
        "BPCompanyRelationship",
        "BPLoanReportingCreditorNumber",
        "BPOeNBIdentNumber",
        "BPOeNBTargetGroup",
        "BPOeNBIdentNumberAssigned",
        "BPOeNBInstituteNumber",
        "BusinessPartnerIsOeNBInstitute",
        "BusinessPartnerGroup",
        "BPGroupAssignmentCategory",
        "BusinessPartnerGroupName",
        "BusinessPartnerLegalEntity",
        "BPGerAstRglnRestrictedAstQuota",
        "BusinessPartnerDebtorGroup",
        "BusinessPartnerBusinessPurpose",
        "BusinessPartnerRiskGroup",
        "BPRiskGroupingDate",
        "BPHasGroupAffiliation",
        "BPIsMonetaryFinInstitution",
        "BPCrdtStandingReviewIsRequired",
        "BPLoanMonitoringIsRequired",
        "BPHasCreditingRelief",
        "BPInvestInRstrcdAstIsAuthzd",
        "BPCentralBankCountryRegion",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_BPFinancialServicesReporting")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class BPFiscalYearInformationTable(CustomAPITable):
    """Business partner fiscal year information of all the available records linked to business partners in the system."""

    name: str = "bp_fiscal_year_information"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "BusinessPartner",
        "BusinessPartnerFiscalYear",
        "BPBalanceSheetCurrency",
        "BPAnnualStockholderMeetingDate",
        "BPFiscalYearStartDate",
        "BPFiscalYearEndDate",
        "BPFiscalYearIsClosed",
        "BPFiscalYearClosingDate",
        "BPFsclYrCnsldtdFinStatementDte",
        "BPCapitalStockAmtInBalShtCrcy",
        "BPIssdStockCptlAmtInBalShtCrcy",
        "BPPartcipnCertAmtInBalShtCrcy",
        "BPEquityCapitalAmtInBalShtCrcy",
        "BPGrossPremiumAmtInBalShtCrcy",
        "BPNetPremiumAmtInBalShtCrcy",
        "BPAnnualSalesAmtInBalShtCrcy",
        "BPAnnualNetIncAmtInBalShtCrcy",
        "BPDividendDistrAmtInBalShtCrcy",
        "BPDebtRatioInYears",
        "BPAnnualPnLAmtInBalShtCrcy",
        "BPBalSheetTotalAmtInBalShtCrcy",
        "BPNumberOfEmployees",
        "BPCptlReserveAmtInBalShtCrcy",
        "BPLglRevnRsrvAmtInBalShtCrcy",
        "RevnRsrvOwnStkAmtInBalShtCrcy",
        "BPStatryReserveAmtInBalShtCrcy",
        "BPOthRevnRsrvAmtInBalShtCrcy",
        "BPPnLCarryfwdAmtInBalShtCrcy",
        "BPSuborddLbltyAmtInBalShtCrcy",
        "BPRetOnTotalCptlEmpldInPercent",
        "BPDebtClearancePeriodInYears",
        "BPFinancingCoeffInPercent",
        "BPEquityRatioInPercent",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_BPFiscalYearInformation")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class BPRelationshipTable(CustomAPITable):
    """Business partner relationship data fields of all the available records in the system"""

    name: str = "bp_relationship"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "RelationshipNumber",
        "BusinessPartner1",
        "BusinessPartner2",
        "ValidityEndDate",
        "ValidityStartDate",
        "IsStandardRelationship",
        "RelationshipCategory",
        "BPRelationshipType",
        "CreatedByUser",
        "CreationDate",
        "CreationTime",
        "LastChangedByUser",
        "LastChangeDate",
        "LastChangeTime",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_BPRelationship")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class BuPaAddressUsageTable(CustomAPITable):
    """All the address usage records linked to all business partner address records in the system"""

    name: str = "bu_pa_address_usage"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "BusinessPartner",
        "ValidityEndDate",
        "AddressUsage",
        "AddressID",
        "ValidityStartDate",
        "StandardUsage",
        "AuthorizationGroup",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_BuPaAddressUsage")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class BuPaIdentificationTable(CustomAPITable):
    """Business partner identification data fields of all the records available records in the system"""

    name: str = "bu_pa_identification"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "BusinessPartner",
        "BPIdentificationType",
        "BPIdentificationNumber",
        "BPIdnNmbrIssuingInstitute",
        "BPIdentificationEntryDate",
        "Country",
        "Region",
        "ValidityStartDate",
        "ValidityEndDate",
        "AuthorizationGroup",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_BuPaIdentification")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class BuPaIndustryTable(CustomAPITable):
    """Business partner industry data fields of all the available records in the system"""

    name: str = "bu_pa_industry"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "IndustrySector",
        "IndustrySystemType",
        "BusinessPartner",
        "IsStandardIndustry",
        "IndustryKeyDescription",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_BuPaIndustry")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class BusinessPartnerTable(CustomAPITable):
    """General data fields of all the business partner records available in the system"""

    name: str = "business_partner"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "BusinessPartner",
        "Customer",
        "Supplier",
        "AcademicTitle",
        "AuthorizationGroup",
        "BusinessPartnerCategory",
        "BusinessPartnerFullName",
        "BusinessPartnerGrouping",
        "BusinessPartnerName",
        "BusinessPartnerUUID",
        "CorrespondenceLanguage",
        "CreatedByUser",
        "CreationDate",
        "CreationTime",
        "FirstName",
        "FormOfAddress",
        "Industry",
        "InternationalLocationNumber1",
        "InternationalLocationNumber2",
        "IsFemale",
        "IsMale",
        "IsNaturalPerson",
        "IsSexUnknown",
        "GenderCodeName",
        "Language",
        "LastChangeDate",
        "LastChangeTime",
        "LastChangedByUser",
        "LastName",
        "LegalForm",
        "OrganizationBPName1",
        "OrganizationBPName2",
        "OrganizationBPName3",
        "OrganizationBPName4",
        "OrganizationFoundationDate",
        "OrganizationLiquidationDate",
        "SearchTerm1",
        "SearchTerm2",
        "AdditionalLastName",
        "BirthDate",
        "BusinessPartnerBirthDateStatus",
        "BusinessPartnerBirthplaceName",
        "BusinessPartnerDeathDate",
        "BusinessPartnerIsBlocked",
        "BusinessPartnerType",
        "ETag",
        "GroupBusinessPartnerName1",
        "GroupBusinessPartnerName2",
        "IndependentAddressID",
        "InternationalLocationNumber3",
        "MiddleName",
        "NameCountry",
        "NameFormat",
        "PersonFullName",
        "PersonNumber",
        "IsMarkedForArchiving",
        "BusinessPartnerIDByExtSystem",
        "BusinessPartnerPrintFormat",
        "BusinessPartnerOccupation",
        "BusPartMaritalStatus",
        "BusPartNationality",
        "BusinessPartnerBirthName",
        "BusinessPartnerSupplementName",
        "NaturalPersonEmployerName",
        "LastNamePrefix",
        "LastNameSecondPrefix",
        "Initials",
        "BPDataControllerIsNotRequired",
        "TradingPartner",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_BusinessPartner")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class BusinessPartnerAddressTable(CustomAPITable):
    """Business partner address data fields of all the available records in the system"""

    name: str = "business_partner_address"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "BusinessPartner",
        "AddressID",
        "ValidityStartDate",
        "ValidityEndDate",
        "AuthorizationGroup",
        "AddressUUID",
        "AdditionalStreetPrefixName",
        "AdditionalStreetSuffixName",
        "AddressTimeZone",
        "CareOfName",
        "CityCode",
        "CityName",
        "CompanyPostalCode",
        "Country",
        "County",
        "DeliveryServiceNumber",
        "DeliveryServiceTypeCode",
        "District",
        "FormOfAddress",
        "FullName",
        "HomeCityName",
        "HouseNumber",
        "HouseNumberSupplementText",
        "Language",
        "POBox",
        "POBoxDeviatingCityName",
        "POBoxDeviatingCountry",
        "POBoxDeviatingRegion",
        "POBoxIsWithoutNumber",
        "POBoxLobbyName",
        "POBoxPostalCode",
        "Person",
        "PostalCode",
        "PrfrdCommMediumType",
        "Region",
        "StreetName",
        "StreetPrefixName",
        "StreetSuffixName",
        "TaxJurisdiction",
        "TransportZone",
        "AddressIDByExternalSystem",
        "CountyCode",
        "TownshipCode",
        "TownshipName",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_BusinessPartnerAddress")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class BusinessPartnerContactTable(CustomAPITable):
    """Business partner contact data fields of all the available records in the system"""

    name: str = "business_partner_contact"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "RelationshipNumber",
        "BusinessPartnerCompany",
        "BusinessPartnerPerson",
        "ValidityEndDate",
        "ValidityStartDate",
        "IsStandardRelationship",
        "RelationshipCategory",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_BusinessPartnerContact")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class BusinessPartnerPaymentCardTable(CustomAPITable):
    """Business partner payment card data fields of all the available records in the system"""

    name: str = "business_partner_payment_card"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "BusinessPartner",
        "PaymentCardID",
        "PaymentCardType",
        "CardNumber",
        "IsStandardCard",
        "CardDescription",
        "ValidityDate",
        "ValidityEndDate",
        "CardHolder",
        "CardIssuingBank",
        "CardIssueDate",
        "PaymentCardLock",
        "MaskedCardNumber",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_BusinessPartnerPaymentCard")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class BusinessPartnerRatingTable(CustomAPITable):
    """Business partner ratings of all the available records linked to business partners in the system"""

    name: str = "business_partner_rating"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "BusinessPartner",
        "BusinessPartnerRatingProcedure",
        "BPRatingValidityEndDate",
        "BusinessPartnerRatingGrade",
        "BusinessPartnerRatingTrend",
        "BPRatingValidityStartDate",
        "BPRatingCreationDate",
        "BusinessPartnerRatingComment",
        "BusinessPartnerRatingIsAllowed",
        "BPRatingIsValidOnKeyDate",
        "BusinessPartnerRatingKeyDate",
        "BusinessPartnerRatingIsExpired",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_BusinessPartnerRating")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class BusinessPartnerRoleTable(CustomAPITable):
    """Business partner role data fields of all the records available records in the system"""

    name: str = "business_partner_role"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "BusinessPartner",
        "BusinessPartnerRole",
        "ValidFrom",
        "ValidTo",
        "AuthorizationGroup",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_BusinessPartnerRole")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class BusinessPartnerTaxNumberTable(CustomAPITable):
    """Tax number data of all the available records linked to business partners in the system"""

    name: str = "business_partner_tax_number"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "BusinessPartner",
        "BPTaxType",
        "BPTaxNumber",
        "BPTaxLongNumber",
        "AuthorizationGroup",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_BusinessPartnerTaxNumber")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class BusPartAddrDepdntTaxNumberTable(CustomAPITable):
    """Address dependent tax number data of all the available records linked to business partners in the system"""

    name: str = "business_partner_address_dependent_tax_number"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "BusinessPartner",
        "AddressID",
        "BPTaxType",
        "BPTaxNumber",
        "BPTaxLongNumber",
        "AuthorizationGroup",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_BusPartAddrDepdntTaxNmbr")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class CustAddrDepdntExtIdentifierTable(CustomAPITable):
    """Address dependent external identifiers of all the available records linked to customers in the system"""

    name: str = "cust_addr_depdnt_ext_identifier"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "Customer",
        "AddressID",
        "CustomerExternalRefID",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_CustAddrDepdntExtIdentifier")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class CustAddrDepdntInformationTable(CustomAPITable):
    """General data of all the customer records available in the system"""

    name: str = "customer"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "Customer",
        "AuthorizationGroup",
        "BillingIsBlockedForCustomer",
        "CreatedByUser",
        "CreationDate",
        "CustomerAccountGroup",
        "CustomerClassification",
        "CustomerFullName",
        "BPCustomerFullName",
        "CustomerName",
        "BPCustomerName",
        "DeliveryIsBlocked",
        "FreeDefinedAttribute01",
        "FreeDefinedAttribute02",
        "FreeDefinedAttribute03",
        "FreeDefinedAttribute04",
        "FreeDefinedAttribute05",
        "FreeDefinedAttribute06",
        "FreeDefinedAttribute07",
        "FreeDefinedAttribute08",
        "FreeDefinedAttribute09",
        "FreeDefinedAttribute10",
        "NFPartnerIsNaturalPerson",
        "OrderIsBlockedForCustomer",
        "PostingIsBlocked",
        "Supplier",
        "CustomerCorporateGroup",
        "FiscalAddress",
        "Industry",
        "IndustryCode1",
        "IndustryCode2",
        "IndustryCode3",
        "IndustryCode4",
        "IndustryCode5",
        "InternationalLocationNumber1",
        "InternationalLocationNumber2",
        "InternationalLocationNumber3",
        "NielsenRegion",
        "PaymentReason",
        "ResponsibleType",
        "TaxNumber1",
        "TaxNumber2",
        "TaxNumber3",
        "TaxNumber4",
        "TaxNumber5",
        "TaxNumberType",
        "VATRegistration",
        "DeletionIndicator",
        "ExpressTrainStationName",
        "TrainStationName",
        "CityCode",
        "County",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_Customer")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class CustomerCompanyTable(CustomAPITable):
    """Customer company data fields of all the available records in the system linked to customer"""

    name: str = "customer_company"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "Customer",
        "CompanyCode",
        "APARToleranceGroup",
        "AccountByCustomer",
        "AccountingClerk",
        "AccountingClerkFaxNumber",
        "AccountingClerkInternetAddress",
        "AccountingClerkPhoneNumber",
        "AlternativePayerAccount",
        "AuthorizationGroup",
        "CollectiveInvoiceVariant",
        "CustomerAccountNote",
        "CustomerHeadOffice",
        "CustomerSupplierClearingIsUsed",
        "HouseBank",
        "InterestCalculationCode",
        "InterestCalculationDate",
        "IntrstCalcFrequencyInMonths",
        "IsToBeLocallyProcessed",
        "ItemIsToBePaidSeparately",
        "LayoutSortingRule",
        "PaymentBlockingReason",
        "PaymentMethodsList",
        "PaymentReason",
        "PaymentTerms",
        "PaytAdviceIsSentbyEDI",
        "PhysicalInventoryBlockInd",
        "ReconciliationAccount",
        "RecordPaymentHistoryIndicator",
        "UserAtCustomer",
        "DeletionIndicator",
        "CashPlanningGroup",
        "KnownOrNegotiatedLeave",
        "ValueAdjustmentKey",
        "CustomerAccountGroup",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_CustomerCompany")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class CustomerCompanyTextTable(CustomAPITable):
    """Customer company text records attached to customer company in the system"""

    name: str = "customer_company_text"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "Customer",
        "CompanyCode",
        "Language",
        "LongTextID",
        "LongText",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_CustomerCompanyText")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class CustomerDunningTable(CustomAPITable):
    """Dunning records attached to customer company in the system"""

    name: str = "customer_dunning"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "Customer",
        "CompanyCode",
        "DunningArea",
        "DunningBlock",
        "DunningLevel",
        "DunningProcedure",
        "DunningRecipient",
        "LastDunnedOn",
        "LegDunningProcedureOn",
        "DunningClerk",
        "AuthorizationGroup",
        "CustomerAccountGroup",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_CustomerDunning")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class CustomerSalesAreaTable(CustomAPITable):
    """Customer sales area data fields of all the available records in the system"""

    name: str = "customer_sales_area"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "Customer",
        "SalesOrganization",
        "DistributionChannel",
        "Division",
        "AccountByCustomer",
        "AuthorizationGroup",
        "BillingIsBlockedForCustomer",
        "CompleteDeliveryIsDefined",
        "CreditControlArea",
        "Currency",
        "CustIsRlvtForSettlmtMgmt",
        "CustomerABCClassification",
        "CustomerAccountAssignmentGroup",
        "CustomerGroup",
        "CustomerIsRebateRelevant",
        "CustomerPaymentTerms",
        "CustomerPriceGroup",
        "CustomerPricingProcedure",
        "CustProdProposalProcedure",
        "DeliveryIsBlockedForCustomer",
        "DeliveryPriority",
        "IncotermsClassification",
        "IncotermsLocation2",
        "IncotermsVersion",
        "IncotermsLocation1",
        "IncotermsSupChnLoc1AddlUUID",
        "IncotermsSupChnLoc2AddlUUID",
        "IncotermsSupChnDvtgLocAddlUUID",
        "DeletionIndicator",
        "IncotermsTransferLocation",
        "InspSbstHasNoTimeOrQuantity",
        "InvoiceDate",
        "ItemOrderProbabilityInPercent",
        "ManualInvoiceMaintIsRelevant",
        "MaxNmbrOfPartialDelivery",
        "OrderCombinationIsAllowed",
        "OrderIsBlockedForCustomer",
        "OverdelivTolrtdLmtRatioInPct",
        "PartialDeliveryIsAllowed",
        "PriceListType",
        "ProductUnitGroup",
        "ProofOfDeliveryTimeValue",
        "SalesGroup",
        "SalesItemProposal",
        "SalesOffice",
        "ShippingCondition",
        "SlsDocIsRlvtForProofOfDeliv",
        "SlsUnlmtdOvrdelivIsAllwd",
        "SupplyingPlant",
        "SalesDistrict",
        "UnderdelivTolrtdLmtRatioInPct",
        "InvoiceListSchedule",
        "ExchangeRateType",
        "AdditionalCustomerGroup1",
        "AdditionalCustomerGroup2",
        "AdditionalCustomerGroup3",
        "AdditionalCustomerGroup4",
        "AdditionalCustomerGroup5",
        "PaymentGuaranteeProcedure",
        "CustomerAccountGroup",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_CustomerSalesArea")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class CustomerSalesAreaTaxTable(CustomAPITable):
    """Customer sales area tax data fields of all the available records in the system"""

    name: str = "customer_sales_area_tax"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "Customer",
        "SalesOrganization",
        "DistributionChannel",
        "Division",
        "DepartureCountry",
        "CustomerTaxCategory",
        "CustomerTaxClassification",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_CustomerSalesAreaTax")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class CustomerSalesAreaTextTable(CustomAPITable):
    """Customer sales area text fields of all the available records in the system linked to customer sales areas"""

    name: str = "customer_sales_area_text"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "Customer",
        "SalesOrganization",
        "DistributionChannel",
        "Division",
        "Language",
        "LongTextID",
        "LongText",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_CustomerSalesAreaText")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class CustomerTaxGroupingTable(CustomAPITable):
    """Customer tax grouping data attached to a customer in the system"""

    name: str = "customer_tax_grouping"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "Customer",
        "CustomerTaxGroupingCode",
        "CustTaxGrpExemptionCertificate",
        "CustTaxGroupExemptionRate",
        "CustTaxGroupExemptionStartDate",
        "CustTaxGroupExemptionEndDate",
        "CustTaxGroupSubjectedStartDate",
        "CustTaxGroupSubjectedEndDate",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_CustomerTaxGrouping")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class CustomerTextTable(CustomAPITable):
    """Customer text data attached to a customer in the system"""

    name: str = "customer_text"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "Customer",
        "Language",
        "LongTextID",
        "LongText",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_CustomerText")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class CustomerUnloadingPointTable(CustomAPITable):
    """Unloading point data attached to a customer in the system"""

    name: str = "customer_unloading_point"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "Customer",
        "UnloadingPointName",
        "CustomerFactoryCalenderCode",
        "BPGoodsReceivingHoursCode",
        "IsDfltBPUnloadingPoint",
        "MondayMorningOpeningTime",
        "MondayMorningClosingTime",
        "MondayAfternoonOpeningTime",
        "MondayAfternoonClosingTime",
        "TuesdayMorningOpeningTime",
        "TuesdayMorningClosingTime",
        "TuesdayAfternoonOpeningTime",
        "TuesdayAfternoonClosingTime",
        "WednesdayMorningOpeningTime",
        "WednesdayMorningClosingTime",
        "WednesdayAfternoonOpeningTime",
        "WednesdayAfternoonClosingTime",
        "ThursdayMorningOpeningTime",
        "ThursdayMorningClosingTime",
        "ThursdayAfternoonOpeningTime",
        "ThursdayAfternoonClosingTime",
        "FridayMorningOpeningTime",
        "FridayMorningClosingTime",
        "FridayAfternoonOpeningTime",
        "FridayAfternoonClosingTime",
        "SaturdayMorningOpeningTime",
        "SaturdayMorningClosingTime",
        "SaturdayAfternoonOpeningTime",
        "SaturdayAfternoonClosingTime",
        "SundayMorningOpeningTime",
        "SundayMorningClosingTime",
        "SundayAfternoonOpeningTime",
        "SundayAfternoonClosingTime",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_CustomerUnloadingPoint")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class CustomerWithHoldingTaxTable(CustomAPITable):
    """Withholding tax records attached to customer company in the system"""

    name: str = "customer_withholding_tax"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "Customer",
        "CompanyCode",
        "WithholdingTaxType",
        "WithholdingTaxCode",
        "WithholdingTaxAgent",
        "ObligationDateBegin",
        "ObligationDateEnd",
        "WithholdingTaxNumber",
        "WithholdingTaxCertificate",
        "WithholdingTaxExmptPercent",
        "ExemptionDateBegin",
        "ExemptionDateEnd",
        "ExemptionReason",
        "AuthorizationGroup",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_CustomerWithHoldingTax")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class CustSalesPartnerFuncTable(CustomAPITable):
    """Partner function fields of all the available records in the system linked to customer sales areas"""

    name: str = "customer_sales_partner_func"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "Customer",
        "SalesOrganization",
        "DistributionChannel",
        "Division",
        "PartnerCounter",
        "PartnerFunction",
        "BPCustomerNumber",
        "CustomerPartnerDescription",
        "DefaultPartner",
        "Supplier",
        "PersonnelNumber",
        "ContactPerson",
        "AddressID",
        "AuthorizationGroup",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_CustSalesPartnerFunc")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class CustSlsAreaAddrDepdntInfoTable(CustomAPITable):
    """Address dependent customer sales area data fields of all the available records in the system"""

    name: str = "customer_sales_area_addr_depdnt_info"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "Customer",
        "SalesOrganization",
        "DistributionChannel",
        "Division",
        "AddressID",
        "IncotermsClassification",
        "IncotermsLocation1",
        "IncotermsLocation2",
        "IncotermsSupChnLoc1AddlUUID",
        "IncotermsSupChnLoc2AddlUUID",
        "IncotermsSupChnDvtgLocAddlUUID",
        "DeliveryIsBlocked",
        "SalesOffice",
        "SalesGroup",
        "ShippingCondition",
        "SupplyingPlant",
        "IncotermsVersion",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_CustSlsAreaAddrDepdntInfo")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class CustSlsAreaAddrDepdntTaxInfoTable(CustomAPITable):
    """Address dependent customer sales area tax data fields of all the available records in the system"""

    name: str = "customer_sales_area_addr_depdnt_tax_info"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "Customer",
        "SalesOrganization",
        "DistributionChannel",
        "Division",
        "AddressID",
        "DepartureCountry",
        "CustomerTaxCategory",
        "CustomerTaxClassification",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_CustSlsAreaAddrDepdntTaxInfo")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class CustUnldgPtAddrDepdntInfoTable(CustomAPITable):
    """Address dependent customer unloading point data fields of all the available records in the system"""

    name: str = "customer_unloading_point_addr_depdnt_info"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "Customer",
        "AddressID",
        "UnloadingPointName",
        "CustomerFactoryCalenderCode",
        "BPGoodsReceivingHoursCode",
        "IsDfltBPUnloadingPoint",
        "MondayMorningOpeningTime",
        "MondayMorningClosingTime",
        "MondayAfternoonOpeningTime",
        "MondayAfternoonClosingTime",
        "TuesdayMorningOpeningTime",
        "TuesdayMorningClosingTime",
        "TuesdayAfternoonOpeningTime",
        "TuesdayAfternoonClosingTime",
        "WednesdayMorningOpeningTime",
        "WednesdayMorningClosingTime",
        "WednesdayAfternoonOpeningTime",
        "WednesdayAfternoonClosingTime",
        "ThursdayMorningOpeningTime",
        "ThursdayMorningClosingTime",
        "ThursdayAfternoonOpeningTime",
        "ThursdayAfternoonClosingTime",
        "FridayMorningOpeningTime",
        "FridayMorningClosingTime",
        "FridayAfternoonOpeningTime",
        "FridayAfternoonClosingTime",
        "SaturdayMorningOpeningTime",
        "SaturdayMorningClosingTime",
        "SaturdayAfternoonOpeningTime",
        "SaturdayAfternoonClosingTime",
        "SundayMorningOpeningTime",
        "SundayMorningClosingTime",
        "SundayAfternoonOpeningTime",
        "SundayAfternoonClosingTime",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_CustUnldgPtAddrDepdntInfo")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class SupplierTable(CustomAPITable):
    """General data of all the supplier records available in the system"""

    name: str = "supplier"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "Supplier",
        "AlternativePayeeAccountNumber",
        "AuthorizationGroup",
        "CreatedByUser",
        "CreationDate",
        "Customer",
        "PaymentIsBlockedForSupplier",
        "PostingIsBlocked",
        "PurchasingIsBlocked",
        "SupplierAccountGroup",
        "SupplierFullName",
        "SupplierName",
        "VATRegistration",
        "BirthDate",
        "ConcatenatedInternationalLocNo",
        "DeletionIndicator",
        "FiscalAddress",
        "Industry",
        "InternationalLocationNumber1",
        "InternationalLocationNumber2",
        "InternationalLocationNumber3",
        "IsNaturalPerson",
        "PaymentReason",
        "ResponsibleType",
        "SuplrQltyInProcmtCertfnValidTo",
        "SuplrQualityManagementSystem",
        "SupplierCorporateGroup",
        "SupplierProcurementBlock",
        "TaxNumber1",
        "TaxNumber2",
        "TaxNumber3",
        "TaxNumber4",
        "TaxNumber5",
        "TaxNumberResponsible",
        "TaxNumberType",
        "SuplrProofOfDelivRlvtCode",
        "BR_TaxIsSplit",
        "DataExchangeInstructionKey",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_Supplier")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class SupplierCompanyTable(CustomAPITable):
    """supplier company data available in the system"""

    name: str = "supplier_company"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "Supplier",
        "CompanyCode",
        "AuthorizationGroup",
        "CompanyCodeName",
        "PaymentBlockingReason",
        "SupplierIsBlockedForPosting",
        "AccountingClerk",
        "AccountingClerkFaxNumber",
        "AccountingClerkPhoneNumber",
        "SupplierClerk",
        "SupplierClerkURL",
        "PaymentMethodsList",
        "PaymentReason",
        "PaymentTerms",
        "ClearCustomerSupplier",
        "IsToBeLocallyProcessed",
        "ItemIsToBePaidSeparately",
        "PaymentIsToBeSentByEDI",
        "HouseBank",
        "CheckPaidDurationInDays",
        "Currency",
        "BillOfExchLmtAmtInCoCodeCrcy",
        "SupplierClerkIDBySupplier",
        "ReconciliationAccount",
        "InterestCalculationCode",
        "InterestCalculationDate",
        "IntrstCalcFrequencyInMonths",
        "SupplierHeadOffice",
        "AlternativePayee",
        "LayoutSortingRule",
        "APARToleranceGroup",
        "SupplierCertificationDate",
        "SupplierAccountNote",
        "WithholdingTaxCountry",
        "DeletionIndicator",
        "CashPlanningGroup",
        "IsToBeCheckedForDuplicates",
        "MinorityGroup",
        "SupplierAccountGroup",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_SupplierCompany")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class SupplierCompanyTextTable(CustomAPITable):
    """Supplier company text data attached to supplier company in the system"""

    name: str = "supplier_company_text"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "Supplier",
        "CompanyCode",
        "Language",
        "LongTextID",
        "LongText",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_SupplierCompanyText")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class SupplierDunningTable(CustomAPITable):
    """Dunning records attached to supplier company in the system"""

    name: str = "supplier_dunning"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "Supplier",
        "CompanyCode",
        "DunningArea",
        "DunningBlock",
        "DunningLevel",
        "DunningProcedure",
        "DunningRecipient",
        "LastDunnedOn",
        "LegDunningProcedureOn",
        "DunningClerk",
        "AuthorizationGroup",
        "SupplierAccountGroup",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_SupplierDunning")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class SupplierPartnerFuncTable(CustomAPITable):
    """Partner function fields of all the available records in the system linked to supplier purchasing organization"""

    name: str = "supplier_partner_func"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "Supplier",
        "PurchasingOrganization",
        "SupplierSubrange",
        "Plant",
        "PartnerFunction",
        "PartnerCounter",
        "DefaultPartner",
        "CreationDate",
        "CreatedByUser",
        "ReferenceSupplier",
        "AuthorizationGroup",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_SupplierPartnerFunc")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class SupplierPurchasingOrgTable(CustomAPITable):
    """Supplier purchasing organization data attached to supplier records in the system"""

    name: str = "supplier_purchasing_org"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "Supplier",
        "PurchasingOrganization",
        "AutomaticEvaluatedRcptSettlmt",
        "CalculationSchemaGroupCode",
        "DeletionIndicator",
        "EvaldReceiptSettlementIsActive",
        "IncotermsClassification",
        "IncotermsTransferLocation",
        "IncotermsVersion",
        "IncotermsLocation1",
        "IncotermsLocation2",
        "IncotermsSupChnLoc1AddlUUID",
        "IncotermsSupChnLoc2AddlUUID",
        "IncotermsSupChnDvtgLocAddlUUID",
        "IntrastatCrsBorderTrMode",
        "InvoiceIsGoodsReceiptBased",
        "InvoiceIsMMServiceEntryBased",
        "MaterialPlannedDeliveryDurn",
        "MinimumOrderAmount",
        "PaymentTerms",
        "PlanningCycle",
        "PricingDateControl",
        "ProdStockAndSlsDataTransfPrfl",
        "ProductUnitGroup",
        "PurOrdAutoGenerationIsAllowed",
        "PurchaseOrderCurrency",
        "PurchasingGroup",
        "PurchasingIsBlockedForSupplier",
        "RoundingProfile",
        "ShippingCondition",
        "SuplrDiscountInKindIsGranted",
        "SuplrInvcRevalIsAllowed",
        "SuplrIsRlvtForSettlmtMgmt",
        "SuplrPurgOrgIsRlvtForPriceDetn",
        "SupplierABCClassificationCode",
        "SupplierAccountNumber",
        "SupplierIsReturnsSupplier",
        "SupplierPhoneNumber",
        "SupplierRespSalesPersonName",
        "SupplierConfirmationControlKey",
        "IsOrderAcknRqd",
        "AuthorizationGroup",
        "SupplierAccountGroup",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_SupplierPurchasingOrg")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class SupplierPurchasingOrgTextTable(CustomAPITable):
    """Supplier purchasing organization text data attached to purchasing organization in the system"""

    name: str = "supplier_purchasing_org_text"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "Supplier",
        "PurchasingOrganization",
        "Language",
        "LongTextID",
        "LongText",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_SupplierPurchasingOrgText")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class SupplierTextTable(CustomAPITable):
    """Supplier text data attached to purchasing organization in the system"""

    name: str = "supplier_text"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "Supplier",
        "Language",
        "LongTextID",
        "LongText",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_SupplierText")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)


class SupplierWithHoldingTaxTable(CustomAPITable):
    """Withholding tax records attached to supplier company in the system"""

    name: str = "supplier_withholding_tax"
    columns: List[str] = [
        "id",
        "uri",
        "type",
        "Supplier",
        "CompanyCode",
        "WithholdingTaxType",
        "ExemptionDateBegin",
        "ExemptionDateEnd",
        "ExemptionReason",
        "IsWithholdingTaxSubject",
        "RecipientType",
        "WithholdingTaxCertificate",
        "WithholdingTaxCode",
        "WithholdingTaxExmptPercent",
        "WithholdingTaxNumber",
        "AuthorizationGroup",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get("A_SupplierWithHoldingTax")
        df = pd.DataFrame.from_records(data)
        return self.apply_query_params(df, query)



import textwrap
from pydantic import BaseModel, Extra

import google.generativeai as palm
import numpy as np
import pandas as pd

from mindsdb.utilities.hooks import before_palm_query, after_palm_query
from mindsdb.utilities import log
from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.integrations.libs.llm.utils import get_completed_prompts

from mindsdb.integrations.utilities.handler_utils import get_api_key

CHAT_MODELS = (
    "models/chat-bison-001",
    "models/embedding-gecko-001",
    "models/text-bison-001",
)

logger = log.getLogger(__name__)


class PalmHandlerArgs(BaseModel):
    target: str = None
    model_name: str = "models/chat-bison-001"
    mode: str = "default"
    predict_params: dict = None
    input_text: str = None
    ft_api_info: dict = None
    ft_result_stats: dict = None
    runtime: str = None
    max_output_tokens: int = 64
    temperature: float = 0.0
    api_key: str = None
    palm_api_key: str = None

    question_column: str = None
    answer_column: str = None
    context_column: str = None
    prompt_template: str = None
    prompt: str = None
    user_column: str = None
    assistant_column: str = None

    class Config:
        # for all args that are not expected, raise an error
        extra = Extra.forbid


class PalmHandler(BaseMLEngine):
    name = "palm"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generative = True
        self.model_name = "models/chat-bison-001"
        self.model_name = (
            "default"  # can also be 'conversational' or 'conversational-full'
        )
        self.supported_modes = [
            "default",
            "conversational",
            "conversational-full",
            "embedding",
        ]
        self.rate_limit = 60  # requests per minute
        self.max_batch_size = 20
        self.default_max_output_tokens = 64
        self.chat_completion_models = CHAT_MODELS

    @staticmethod
    def create_validation(target, args=None, **kwargs):
        if "using" not in args:
            raise Exception(
                "palm engine requires a USING clause! Refer to its documentation for more details."
            )
        else:
            args = args["using"]

        args_model = PalmHandlerArgs(**args)

        if (
            len(set(args.keys()) & {"question_column", "prompt_template", "prompt"})
            == 0
        ):
            raise Exception(
                "One of `question_column` or `prompt_template` is required for this engine."
            )

        # TODO: add example_column for conversational mode
        keys_collection = [
            ["prompt_template"],
            ["question_column", "context_column"],
            ["prompt", "user_column", "assistant_column"],
        ]
        for keys in keys_collection:
            if keys[0] in args and any(
                x[0] in args for x in keys_collection if x != keys
            ):
                raise Exception(
                    textwrap.dedent(
                        """\
                    Please provide one of
                        1) a `prompt_template`
                        2) a `question_column` and an optional `context_column`
                        3) a `prompt' and 'user_column' and 'assistant_column`
                """
                    )
                )

    def create(self, target, args=None, **kwargs):
        args = args["using"]
        args_model = PalmHandlerArgs(**args)

        args_model.target = target
        api_key = get_api_key("palm", args["using"], self.engine_storage, strict=False)

        # Set palm api key
        palm.configure(api_key=api_key)

        available_models = [m.name for m in palm.list_models()]

        if not args_model.model_name:
            args_model.model_name = self.model_name
        elif args_model.model_name not in available_models:
            raise Exception(f"Invalid model name. Please use one of {available_models}")

        if not args_model.mode:
            args_model.mode = self.model_name
        elif args_model.mode not in self.supported_modes:
            raise Exception(
                f"Invalid operation mode. Please use one of {self.supported_modes}"
            )

        self.model_storage.json_set("args", args_model.model_dump())

    def predict(self, df, args=None):
        """
        If there is a prompt template, we use it. Otherwise, we use the concatenation of `context_column` (optional) and `question_column` to ask for a completion.
        """  # noqa
        # TODO: support for edits, embeddings and moderation

        pred_args = args["predict_params"] if args else {}
        args_model = PalmHandlerArgs(**self.model_storage.json_get("args"))
        df = df.reset_index(drop=True)

        if pred_args.get("mode"):
            if pred_args["mode"] in self.supported_modes:
                args_model.mode = pred_args["mode"]
            else:
                raise Exception(
                    f"Invalid operation mode. Please use one of {self.supported_modes}."
                )  # noqa

        if pred_args.get("prompt_template", False):
            base_template = pred_args[
                "prompt_template"
            ]  # override with predict-time template if available
        elif args_model.prompt_template:
            base_template = args_model.prompt_template
        else:
            base_template = None

        # Embedding Mode
        if args_model.mode == "embedding":
            api_args = {
                "model": pred_args.get("model_name", "models/embedding-gecko-001")
            }
            model_name = "models/embedding-gecko-001"
            if args_model.question_column:
                prompts = list(df[args_model.question_column].apply(lambda x: str(x)))
                empty_prompt_ids = np.where(
                    df[[args_model.question_column]].isna().all(axis=1).values
                )[0]
            else:
                raise Exception("Embedding mode needs a question_column")

        # Chat or normal completion mode
        else:
            if (
                args_model.question_column
                and args_model.question_column not in df.columns
            ):
                raise Exception(
                    f"This model expects a question to answer in the '{args_model.question_column}' column."
                )

            if (
                args_model.context_column
                and args_model.context_column not in df.columns
            ):
                raise Exception(
                    f"This model expects context in the '{args_model.context_column}' column."
                )

            # api argument validation
            model_name = args_model.model_name
            api_args = {
                "max_output_tokens": pred_args.get(
                    "max_output_tokens",
                    args_model.max_output_tokens,
                ),
                "temperature": min(
                    1.0,
                    max(0.0, pred_args.get("temperature", args_model.temperature)),
                ),
                "top_p": pred_args.get("top_p", None),
                "candidate_count": pred_args.get("candidate_count", None),
                "stop_sequences": pred_args.get("stop_sequences", None),
            }

            if (
                args_model.mode != "default"
                and model_name not in self.chat_completion_models
            ):
                raise Exception(
                    f"Conversational modes are only available for the following models: {', '.join(self.chat_completion_models)}"
                )  # noqa

            if args_model.prompt_template:
                prompts, empty_prompt_ids = get_completed_prompts(
                    base_template, df
                )
                if len(prompts) == 0:
                    raise Exception("No prompts found")

            elif args_model.context_column:
                empty_prompt_ids = np.where(
                    df[[args_model.context_column, args_model.question_column]]
                    .isna()
                    .all(axis=1)
                    .values
                )[0]
                contexts = list(df[args_model.context_column].apply(lambda x: str(x)))
                questions = list(df[args_model.question_column].apply(lambda x: str(x)))
                prompts = [
                    f"Context: {c}\nQuestion: {q}\nAnswer: "
                    for c, q in zip(contexts, questions)
                ]
                api_args["context"] = "".join(contexts)

            elif args_model.prompt:
                empty_prompt_ids = []
                prompts = list(df[args_model.user_column])
                if len(prompts) == 0:
                    raise Exception("No prompts found")
            else:
                empty_prompt_ids = np.where(
                    df[[args_model.question_column]].isna().all(axis=1).values
                )[0]
                prompts = list(df[args_model.question_column].apply(lambda x: str(x)))

        # remove prompts without signal from completion queue
        prompts = [j for i, j in enumerate(prompts) if i not in empty_prompt_ids]

        api_key = get_api_key("palm", args["using"], self.engine_storage, strict=False)
        api_args = {
            k: v for k, v in api_args.items() if v is not None
        }  # filter out non-specified api args
        completion = self._completion(
            model_name, prompts, api_key, api_args, args_model, df
        )

        # add null completion for empty prompts
        for i in sorted(empty_prompt_ids):
            completion.insert(i, None)

        pred_df = pd.DataFrame(completion, columns=[args_model.target])

        return pred_df

    def _completion(self, model_name, prompts, api_key, api_args, args_model, df):
        """
        Handles completion for an arbitrary amount of rows.
        Additionally, single completion calls are done with exponential backoff to guarantee all prompts are processed,
        because even with previous checks the tokens-per-minute limit may apply.
        """

        def _submit_completion(model_name, prompts, api_key, api_args, args_model, df):
            kwargs = {
                "model": model_name,
            }

            # configure the PaLM SDK with the provided API KEY
            palm.configure(api_key=api_key)

            if model_name == "models/embedding-gecko-001":
                prompts = "".join(prompts)
                return _submit_embedding_completion(kwargs, prompts, api_args)
            elif model_name == args_model.model_name:
                return _submit_chat_completion(
                    kwargs,
                    prompts,
                    api_args,
                    df,
                    mode=args_model.mode,
                )
            else:
                prompts = "".join(prompts)
                return _submit_normal_completion(kwargs, prompts, api_args)

        def _log_api_call(params, response):
            after_palm_query(params, response)

            params2 = params.copy()
            params2.pop("palm_api_key", None)
            params2.pop("user", None)
            logger.debug(f">>>palm call: {params2}:\n{response}")

        def _submit_normal_completion(kwargs, prompts, api_args):
            def _tidy(comp):
                tidy_comps = []
                if comp.candidates and len(comp.candidates) == 0:
                    return ["No completions found"]
                for c in comp.candidates:
                    if "output" in c:
                        tidy_comps.append(c["output"].strip("\n").strip(""))
                return tidy_comps

            kwargs["prompt"] = prompts
            kwargs = {**kwargs, **api_args}

            before_palm_query(kwargs)

            # call the palm sdk with text-bison-001 model
            resp = _tidy(palm.generate_text(**kwargs))
            _log_api_call(kwargs, resp)
            return resp

        def _submit_embedding_completion(kwargs, prompts, api_args):
            def _tidy(comp):
                tidy_comps = []
                if not "embedding" in comp:
                    return [f"No completion found, err {comp}"]
                for c in comp["embedding"]:
                    tidy_comps.append([c])
                return tidy_comps

            kwargs = {}
            kwargs["model"] = api_args["model"]
            kwargs["text"] = prompts

            before_palm_query(kwargs)

            # call the palm sdk with embedding-gecko-001 model
            resp = _tidy(palm.generate_embeddings(**kwargs))
            _log_api_call(kwargs, resp)
            return resp

        def _submit_chat_completion(
            kwargs, prompts, api_args, df, mode="conversational"
        ):
            def _tidy(comp):
                tidy_comps = []
                if comp.candidates and len(comp.candidates) == 0:
                    return ["No completions found"]

                for c in comp.candidates:
                    if "content" in c:
                        tidy_comps.append(c["content"].strip("\n").strip(""))
                    if "output" in c:
                        tidy_comps.append(c["output"].strip("\n").strip(""))
                return tidy_comps

            completions = []
            if mode != "conversational":
                initial_prompt = {
                    "author": "system",
                    "content": "You are a helpful assistant. Your task is to continue the chat.",
                }  # noqa
            else:
                # get prompt from model
                prompt = "".join(prompts)
                initial_prompt = {"author": "system", "content": prompt}  # noqa
                kwargs["messages"] = [initial_prompt]

            last_completion_content = None

            for pidx in range(len(prompts)):
                if mode == "conversational":
                    kwargs["messages"].append(
                        {"author": "user", "content": prompts[pidx]}
                    )

                if mode == "conversational-full" or (
                    mode == "conversational" and pidx == len(prompts) - 1
                ):
                    pkwargs = {**kwargs, **api_args}
                    pkwargs["candidate_count"] = 3
                    pkwargs.pop("max_output_tokens")
                    before_palm_query(kwargs)

                    # call the palm sdk with chat-bison-001 model
                    resp = _tidy(palm.chat(**pkwargs))

                    _log_api_call(pkwargs, resp)

                    completions.extend(resp)
                elif mode == "default":
                    pkwargs = {**kwargs, **api_args}

                    pkwargs["model"] = "models/text-bison-001"
                    pkwargs["prompt"] = prompts[pidx]
                    before_palm_query(kwargs)
                    if pkwargs["prompt"] == "":
                        return ["No prompt provided"]

                    # call the palm sdk with text-bison-001 model
                    resp = _tidy(palm.generate_text(**pkwargs))
                    _log_api_call(pkwargs, resp)

                    completions.extend(resp)
                else:
                    # in "normal" conversational mode, we request completions only for the last row
                    last_completion_content = None
                    if args_model.answer_column in df.columns:
                        # insert completion if provided, which saves redundant API calls
                        completions.extend([df.iloc[pidx][args_model.answer_column]])
                    else:
                        completions.extend([""])

                if args_model.answer_column in df.columns:
                    kwargs["messages"].append(
                        {
                            "author": "assistant",
                            "content": df.iloc[pidx][args_model.answer_column],
                        }
                    )
                elif last_completion_content:
                    # interleave assistant responses with user input
                    kwargs["messages"].append(
                        {"author": "assistant", "content": last_completion_content[0]}
                    )

            return completions

        try:
            completion = _submit_completion(
                model_name, prompts, api_key, api_args, args_model, df
            )
            return completion
        except Exception as e:
            completion = []
            logger.exception(e)
            completion.extend({"error": str(e)})

        return completion



from mindsdb.integrations.libs.const import HANDLER_TYPE

from .__about__ import __version__ as version, __description__ as description
from .connection_args import connection_args, connection_args_example

try:
    from .confluence_handler import ConfluenceHandler as Handler
    import_error = None
except Exception as e:
    Handler = None
    import_error = e

title = "Confluence"
name = "confluence"
type = HANDLER_TYPE.DATA
icon_path = "icon.svg"

__all__ = [
    "Handler",
    "version",
    "name",
    "type",
    "title",
    "description",
    "import_error",
    "icon_path",
    "connection_args_example",
    "connection_args",
]



from .apply_predictor_step import ApplyPredictorStepCall, ApplyPredictorRowStepCall, ApplyTimeseriesPredictorStepCall
from .delete_step import DeleteStepCall
from .fetch_dataframe import FetchDataframeStepCall
from .insert_step import InsertToTableCall, SaveToTableCall, CreateTableCall
from .join_step import JoinStepCall
from .map_reduce_step import MapReduceStepCall
from .multiple_step import MultipleStepsCall
from .prepare_steps import GetPredictorColumnsCall, GetTableColumnsCall
from .project_step import ProjectStepCall
from .sql_steps import LimitOffsetStepCall, DataStepCall
from .subselect_step import SubSelectStepCall, QueryStepCall
from .union_step import UnionStepCall
from .update_step import UpdateToTableCall



from typing import Optional

import dill
import dask
import pandas as pd
from ludwig.automl import auto_train

from mindsdb.integrations.libs.base import BaseMLEngine
from .utils import RayConnection


class LudwigHandler(BaseMLEngine):
    """
    Integration with the Ludwig declarative ML library.
    """  # noqa

    name = 'ludwig'

    def create(self, target: str, df: Optional[pd.DataFrame] = None, args: Optional[dict] = None) -> None:
        args = args['using']  # ignore the rest of the problem definition

        # TODO: filter out incompatible use cases (e.g. time series won't work currently)
        # TODO: enable custom values via `args` (mindful of local vs cloud)
        user_config = {'hyperopt': {'executor': {'gpu_resources_per_trial': 0, 'num_samples': 3}}}  # no GPU for now

        with RayConnection():
            results = auto_train(
                dataset=df,
                target=target,
                tune_for_memory=False,
                time_limit_s=120,
                user_config=user_config,
                # output_directory='./',
                # random_seed=42,
                # use_reference_config=False,
                # kwargs={}
            )
        model = results.best_model
        args['dtype_dict'] = {f['name']: f['type'] for f in model.base_config['input_features']}
        args['accuracies'] = {'metric': results.experiment_analysis.best_result['metric_score']}
        self.model_storage.json_set('args', args)
        self.model_storage.file_set('model', dill.dumps(model))

    def predict(self, df, args=None):
        model = dill.loads(self.model_storage.file_get('model'))
        with RayConnection():
            predictions = self._call_model(df, model)
        return predictions

    @staticmethod
    def _call_model(df, model):
        predictions = dask.compute(model.predict(df)[0])[0]
        target_name = model.config['output_features'][0]['column']

        if target_name not in df:
            predictions.columns = [target_name]
        else:
            predictions.columns = ['prediction']

        predictions[f'{target_name}_explain'] = None
        joined = df.join(predictions)

        if 'prediction' in joined:
            joined = joined.rename({
                target_name: f'{target_name}_original',
                'prediction': target_name
            }, axis=1)
        return joined



import uuid

from bson.int64 import Int64
from mindsdb_sql_parser.ast import Show, Identifier

from mindsdb.api.mongo.classes import Responder
import mindsdb.api.mongo.functions as helpers
from mindsdb.api.mongo.classes.query_sql import run_sql_command


class Responce(Responder):
    when = {'listCollections': helpers.is_true}

    def result(self, query, request_env, mindsdb_env, session):
        database = request_env['database']
        ast_query = Show(
            category='tables',
            from_table=Identifier(parts=[database])
        )
        data = run_sql_command(request_env, ast_query)

        tables = []
        for row in data:
            name = list(row.values())[0]  # first value
            tables.append({
                'name': name,
                'type': 'collection',
                'options': {},
                'info': {
                    'readOnly': False,
                    'uuid': uuid.uuid3(uuid.NAMESPACE_DNS, name).bytes,
                },
                'idIndex': {
                    'v': 2,
                    'key': {
                        '_id': 1
                    },
                    'name': '_id_',
                    'ns': f'{database}.{name}'
                }
            })

        cursor = {
            'id': Int64(0),  # should we save id somewhere?
            'ns': f'{database}.$cmd.listCollections',
            'firstBatch': tables
        }

        return {
            'cursor': cursor,
            'ok': 1,
        }


responder = Responce()



from urllib.parse import quote

import pandas as pd
from sqlalchemy import create_engine
from sqlalchemy.exc import SQLAlchemyError
from clickhouse_sqlalchemy.drivers.base import ClickHouseDialect
from mindsdb_sql_parser.ast.base import ASTNode
from mindsdb.utilities.render.sqlalchemy_render import SqlalchemyRender

from mindsdb.utilities import log
from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)

logger = log.getLogger(__name__)


class ClickHouseHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the ClickHouse statements.
    """

    name = 'clickhouse'

    def __init__(self, name, connection_data, **kwargs):
        super().__init__(name)
        self.dialect = 'clickhouse'
        self.connection_data = connection_data
        self.renderer = SqlalchemyRender(ClickHouseDialect)
        self.is_connected = False
        self.protocol = connection_data.get('protocol', 'native')

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def connect(self):
        """
        Establishes a connection to a ClickHouse server using SQLAlchemy.

        Raises:
            SQLAlchemyError: If an error occurs while connecting to the database.

        Returns:
            Connection: A SQLAlchemy Connection object to the ClickHouse database.
        """
        if self.is_connected:
            return self.connection

        protocol = "clickhouse+native" if self.protocol == 'native' else "clickhouse+http"
        host = quote(self.connection_data['host'])
        port = self.connection_data['port']
        user = quote(self.connection_data['user'])
        password = quote(self.connection_data['password'])
        database = quote(self.connection_data['database'])
        url = f'{protocol}://{user}:{password}@{host}:{port}/{database}'
        # This is not redundunt. Check https://clickhouse-sqlalchemy.readthedocs.io/en/latest/connection.html#http
        if self.protocol == 'https':
            url = url + "?protocol=https"
        try:
            engine = create_engine(url)
            connection = engine.raw_connection()
            self.is_connected = True
            self.connection = connection
        except SQLAlchemyError as e:
            logger.error(f'Error connecting to ClickHouse {self.connection_data["database"]}, {e}!')
            self.is_connected = False
            raise

        return self.connection

    def check_connection(self) -> StatusResponse:
        """
        Checks the status of the connection to the ClickHouse.

        Returns:
            StatusResponse: An object containing the success status and an error message if an error occurs.
        """
        response = StatusResponse(False)
        need_to_close = not self.is_connected

        try:
            connection = self.connect()
            cur = connection.cursor()
            try:
                cur.execute('select 1;')
            finally:
                cur.close()
            response.success = True
        except SQLAlchemyError as e:
            logger.error(f'Error connecting to ClickHouse {self.connection_data["database"]}, {e}!')
            response.error_message = str(e)
            self.is_connected = False

        if response.success is True and need_to_close:
            self.disconnect()

        return response

    def native_query(self, query: str) -> Response:
        """
        Executes a SQL query and returns the result.

        Args:
            query (str): The SQL query to be executed.

        Returns:
            Response: A response object containing the result of the query or an error message.
        """

        connection = self.connect()
        cur = connection.cursor()
        try:
            cur.execute(query)
            result = cur.fetchall()
            if result:
                response = Response(
                    RESPONSE_TYPE.TABLE,
                    pd.DataFrame(
                        result,
                        columns=[x[0] for x in cur.description]
                    )
                )
            else:
                response = Response(RESPONSE_TYPE.OK)
            connection.commit()
        except SQLAlchemyError as e:
            logger.error(f'Error running query: {query} on {self.connection_data["database"]}!')
            response = Response(
                RESPONSE_TYPE.ERROR,
                error_message=str(e)
            )
            connection.rollback()
        finally:
            cur.close()

        return response

    def query(self, query: ASTNode) -> Response:
        """
        Retrieve the data from the SQL statement with eliminated rows that dont satisfy the WHERE condition
        """
        query_str = self.renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)

    def get_tables(self) -> Response:
        """
        Get a list with all of the tabels in ClickHouse db
        """
        q = f"SHOW TABLES FROM {self.connection_data['database']}"
        result = self.native_query(q)
        df = result.data_frame

        if df is not None:
            result.data_frame = df.rename(columns={df.columns[0]: 'table_name'})

        return result

    def get_columns(self, table_name) -> Response:
        """
        Show details about the table
        """
        q = f"DESCRIBE {table_name}"
        result = self.native_query(q)
        return result



import copy
from typing import List

from mindsdb_sql_parser.ast import Identifier, Select, BinaryOperation, Constant, Parameter
from mindsdb_sql_parser import ast

from mindsdb.integrations.utilities.query_traversal import query_traversal
from mindsdb.api.executor.planner.exceptions import PlanningException


def get_predictor_name_identifier(identifier):
    new_identifier = copy.deepcopy(identifier)
    if len(new_identifier.parts) > 1:
        new_identifier.parts.pop(0)
    return new_identifier


def disambiguate_predictor_column_identifier(identifier, predictor):
    """Removes integration name from column if it's present, adds table path if it's absent"""
    table_ref = predictor.alias.parts_to_str() if predictor.alias else predictor.parts_to_str()
    parts = list(identifier.parts)
    if parts[0] == table_ref:
        parts = parts[1:]

    new_identifier = Identifier(parts=parts)
    return new_identifier


def recursively_extract_column_values(op, row_dict, predictor):
    if isinstance(op, BinaryOperation) and op.op == '=':
        id = op.args[0]
        value = op.args[1]

        if not (
            isinstance(id, Identifier)
            and (isinstance(value, Constant) or isinstance(value, Parameter))
        ):
            raise PlanningException(f'The WHERE clause for selecting from a predictor'
                                    f' must contain pairs \'Identifier(...) = Constant(...)\','
                                    f' found instead: {id.to_tree()}, {value.to_tree()}')

        id = disambiguate_predictor_column_identifier(id, predictor)

        if str(id) in row_dict:
            raise PlanningException(f'Multiple values provided for {str(id)}')
        if isinstance(value, Constant):
            value = value.value
        row_dict[str(id)] = value
    elif isinstance(op, BinaryOperation) and op.op == 'and':
        recursively_extract_column_values(op.args[0], row_dict, predictor)
        recursively_extract_column_values(op.args[1], row_dict, predictor)
    else:
        raise PlanningException(f'Only \'and\' and \'=\' operations allowed in WHERE clause, found: {op.to_tree()}')


def get_deepest_select(select):
    if not select.from_table or not isinstance(select.from_table, Select):
        return select
    return get_deepest_select(select.from_table)


def convert_join_to_list(join):
    # join tree to table list

    if isinstance(join.right, ast.Join):
        raise NotImplementedError('Wrong join AST')

    items = []

    if isinstance(join.left, ast.Join):
        # dive to next level
        items.extend(convert_join_to_list(join.left))
    else:
        # this is first table
        items.append(dict(
            table=join.left
        ))

    # all properties set to right table
    items.append(dict(
        table=join.right,
        join_type=join.join_type,
        is_implicit=join.implicit,
        condition=join.condition
    ))

    return items


def get_query_params(query):
    # find all parameters
    params = []

    def params_find(node, **kwargs):
        if isinstance(node, ast.Parameter):
            params.append(node)
            return node

    query_traversal(query, params_find)
    return params


def fill_query_params(query, params):

    params = copy.deepcopy(params)

    def params_replace(node, **kwargs):
        if isinstance(node, ast.Parameter):
            value = params.pop(0)
            return ast.Constant(value)

    # put parameters into query
    query_traversal(query, params_replace)

    return query


def filters_to_bin_op(filters: List[BinaryOperation]):
    # make a new where clause without params
    where = None
    for flt in filters:
        if where is None:
            where = flt
        else:
            where = BinaryOperation(op='and', args=[where, flt])
    return where



from enum import Enum


class PostgresField:
    def __init__(self, name: str, object_id: int, dt_size: int, type_modifier: int, format_code: int, table_id: int = 0,
                 column_id: int = 0):
        self.name = name
        self.object_id = object_id
        self.dt_size = dt_size
        self.type_modifier = type_modifier
        self.format_code = format_code
        self.table_id = table_id
        self.column_id = column_id


class GenericField(PostgresField):
    def __init__(self, name: str, object_id: int, table_id: int = 0, column_id: int = 0):
        super().__init__(name=name, object_id=object_id, dt_size=-1, type_modifier=-1, format_code=0, table_id=table_id,
                         column_id=column_id)


class IntField(PostgresField):
    def __init__(self, name: str, table_id: int = 0, column_id: int = 0):
        super().__init__(name=name, object_id=23, dt_size=4, type_modifier=-1, format_code=0, table_id=table_id,
                         column_id=column_id)


class POSTGRES_TYPES(Enum):
    VARCHAR = 0
    INT = 1
    LONG = 2
    DOUBLE = 3
    DATETIME = 4
    DATE = 5



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    host={
        "type": ARG_TYPE.STR,
        "description": "chromadb server host",
        "required": False,
    },
    port={
        "type": ARG_TYPE.STR,
        "description": "chromadb server port",
        "required": False,
    },
    persist_directory={
        "type": ARG_TYPE.STR,
        "description": "persistence directory for ChromaDB",
        "required": False,
    },
)

connection_args_example = OrderedDict(
    host="localhost",
    port="8000",
    persist_directory="chromadb",
)



from typing import Optional

import pandas as pd
import pinotdb
import requests
from requests.exceptions import InvalidSchema
import json

from mindsdb_sql_parser import parse_sql
from mindsdb.utilities.render.sqlalchemy_render import SqlalchemyRender
from mindsdb.integrations.libs.base import DatabaseHandler
from pinotdb.sqlalchemy import PinotDialect

from mindsdb_sql_parser.ast.base import ASTNode

from mindsdb.utilities import log
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)


logger = log.getLogger(__name__)

class PinotHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the Apache Pinot statements.
    """

    name = 'pinot'

    def __init__(self, name: str, connection_data: Optional[dict], **kwargs):
        """
        Initialize the handler.
        Args:
            name (str): name of particular handler instance
            connection_data (dict): parameters for connecting to the database
            **kwargs: arbitrary keyword arguments.
        """
        super().__init__(name)
        self.parser = parse_sql
        self.dialect = 'pinot'

        optional_parameters = ['username', 'password']
        for parameter in optional_parameters:
            if parameter not in connection_data:
                connection_data[parameter] = None

        if 'verify_ssl' not in connection_data:
            connection_data['verify_ssl'] = 'False'

        if 'scheme' not in connection_data:
            connection_data['scheme'] = 'http'

        self.connection_data = connection_data
        self.kwargs = kwargs

        self.connection = None
        self.is_connected = False

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def connect(self) -> StatusResponse:
        """
        Set up the connection required by the handler.
        Returns:
            HandlerStatusResponse
        """

        if self.is_connected is True:
            return self.connection

        self.connection = pinotdb.connect(
            host=self.connection_data['host'],
            port=self.connection_data['broker_port'],
            path=self.connection_data['path'],
            scheme=self.connection_data['scheme'],
            username=self.connection_data['username'],
            password=self.connection_data['password'],
            verify_ssl=json.loads(self.connection_data['verify_ssl'].lower())
        )
        self.is_connected = True

        return self.connection

    def disconnect(self):
        """ Close any existing connections

        Should switch self.is_connected.
        """
        self.is_connected = False
        return

    def check_connection(self) -> StatusResponse:
        """
        Check connection to the handler.
        Returns:
            HandlerStatusResponse
        """

        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            self.connect()
            response.success = True
        except Exception as e:
            logger.error(f'Error connecting to Pinot, {e}!')
            response.error_message = str(e)
        finally:
            if response.success is True and need_to_close:
                self.disconnect()
            if response.success is False and self.is_connected is True:
                self.is_connected = False

        return response

    def native_query(self, query: str) -> StatusResponse:
        """
        Receive raw query and act upon it somehow.
        Args:
            query (str): query in native format
        Returns:
            HandlerResponse
        """

        need_to_close = self.is_connected is False

        connection = self.connect()
        cursor = connection.cursor()

        try:
            cursor.execute(query)
            result = cursor.fetchall()
            if result:
                response = Response(
                    RESPONSE_TYPE.TABLE,
                    data_frame=pd.DataFrame(
                        result,
                        columns=[x[0] for x in cursor.description]
                    )
                )
            else:
                connection.commit()
                response = Response(RESPONSE_TYPE.OK)
        except Exception as e:
            logger.error(f'Error running query: {query} on Pinot!')
            response = Response(
                RESPONSE_TYPE.ERROR,
                error_message=str(e)
            )

        cursor.close()
        if need_to_close is True:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> StatusResponse:
        """
        Receive query as AST (abstract syntax tree) and act upon it somehow.
        Args:
            query (ASTNode): sql query represented as AST. May be any kind
                of query: SELECT, INTSERT, DELETE, etc
        Returns:
            HandlerResponse
        """
        renderer = SqlalchemyRender(PinotDialect)
        query_str = renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)

    def get_tables(self) -> StatusResponse:
        """
        Return list of entities that will be accessible as tables.
        Returns:
            HandlerResponse
        """

        api_url = f"{self.connection_data['host']}:{self.connection_data['controller_port']}/tables"
        try:
            result = requests.get(api_url)
        except InvalidSchema as e:
            api_url = f"{self.connection_data['scheme']}://{api_url}"
            result = requests.get(api_url)

        response = Response(
            RESPONSE_TYPE.TABLE,
            data_frame=pd.DataFrame(
                json.loads(result.content)['tables'],
                columns=['table_name']
            )
        )

        return response

    def get_columns(self, table_name: str) -> StatusResponse:
        """
        Returns a list of entity columns.
        Args:
            table_name (str): name of one of tables returned by self.get_tables()
        Returns:
            HandlerResponse
        """

        api_url = f"{self.connection_data['host']}:{self.connection_data['controller_port']}/tables/{table_name}/schema"
        try:
            result = requests.get(api_url)
        except InvalidSchema as e:
            api_url = f"{self.connection_data['scheme']}://{api_url}"
            result = requests.get(api_url)

        df = pd.DataFrame(json.loads(result.content)['dimensionFieldSpecs'])
        df = df.rename(columns={'name': 'column_name', 'dataType': 'data_type'})

        response = Response(
            RESPONSE_TYPE.TABLE,
            data_frame=df
        )

        return response



from typing import Optional, Dict

import unify
import pandas as pd

from mindsdb.integrations.libs.base import BaseMLEngine

from mindsdb.utilities import log

from mindsdb.integrations.utilities.handler_utils import get_api_key


logger = log.getLogger(__name__)

class UnifyHandler(BaseMLEngine):
    """
    Integration with the Unifyai Python Library
    """
    name = 'unify'

    def create(self, target: str, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:
        if 'using' not in args:
            raise Exception("Unify engine requires a USING clause! Refer to its documentation for more details.")

        self.generative = True
        self.model_storage.json_set('args', args)

    def predict(self, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:
        
        args = self.model_storage.json_get('args')
        

        input_keys = list(args.keys())

        logger.info(f"Input keys: {input_keys}!")

        if 'model' not in args['using']:
            raise Exception("Unify requires an model parameter in the USING clause! Refer to its documentation for more details.")
        model = args['using']['model']

        if 'provider' not in args['using']:
            raise Exception("Unify requires a provider parameter in the USING clause! Refer to its documentation for more details.")
        provider = args['using']['provider']
        
        self.endpoint = model + '@' + provider
        self.api_key = get_api_key('unify', args["using"], self.engine_storage, strict=False)
        available_endpoints = unify.utils.list_endpoints(api_key=self.api_key)
        if self.endpoint not in available_endpoints:
            raise Exception("The model, provider or their combination is not supported by Unify! The supported endpoints are: " + str(available_endpoints))

        question_column = args['using']['column']
        if question_column not in df.columns:
            raise RuntimeError(f'Column "{question_column}" not found in input data')
        
        result_df = pd.DataFrame() 
        result_df['predictions'] = df[question_column].apply(self.predict_answer)     
        result_df = result_df.rename(columns={'predictions': args['target']})
        
        return result_df
    
    def predict_answer(self, text):
        
        client = unify.Unify(self.endpoint, api_key=self.api_key)

        response = client.generate(text)

        return response



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    host={
        'type': ARG_TYPE.STR,
        'description': 'The host name or IP address of the Apache Pinot cluster.'
    },
    broker_port={
        'type': ARG_TYPE.INT,
        'description': 'The port that the Broker of the Apache Pinot cluster is running on.'
    },
    controller_port={
        'type': ARG_TYPE.INT,
        'description': 'The port that the Controller of the Apache Pinot cluster is running on.'
    },
    path={
        'type': ARG_TYPE.STR,
        'description': 'The query path.'
    },
    scheme={
        'type': ARG_TYPE.STR,
        'description': 'The URI schema. This parameter is optional and the default will be https.'
    },
    username={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with the Apache Pinot cluster. This parameter is optional.'
    },
    password={
        'type': ARG_TYPE.PWD,
        'description': 'The password used to authenticate with the Apache Pinot cluster. This parameter is optional.',
        'secret': True
    },
    verify_ssl={
        'type': ARG_TYPE.STR,
        'description': 'The flag for whether SSL certificates should be verified or not. This parameter is optional and '
                       'if specified, it should be either True or False'
    },
)

connection_args_example = OrderedDict(
    host='localhost',
    broker_port=8000,
    controller_port=9000,
    path='/query/sql',
    scheme='http'
)



import os
import importlib
import traceback
import datetime as dt

from sqlalchemy.orm.attributes import flag_modified

from mindsdb_sql_parser import parse_sql
from mindsdb_sql_parser.ast import Identifier, Select, Star, NativeQuery

from mindsdb.api.executor import SQLQuery
import mindsdb.utilities.profiler as profiler
from mindsdb.utilities.functions import mark_process
from mindsdb.utilities.config import Config
from mindsdb.utilities.context import context as ctx
from mindsdb.utilities import log
import mindsdb.interfaces.storage.db as db
from mindsdb.interfaces.storage.model_fs import ModelStorage, HandlerStorage
from mindsdb.interfaces.model.functions import get_model_records
from mindsdb.integrations.utilities.utils import format_exception_error
from mindsdb.integrations.utilities.sql_utils import make_sql_session
from mindsdb.integrations.libs.const import PREDICTOR_STATUS
from mindsdb.integrations.libs.ml_handler_process.handlers_cacher import handlers_cacher

logger = log.getLogger(__name__)


@mark_process(name='learn')
def learn_process(data_integration_ref: dict, problem_definition: dict, fetch_data_query: str,
                  project_name: str, model_id: int, integration_id: int, base_model_id: int,
                  set_active: bool, module_path: str):
    ctx.profiling = {
        'level': 0,
        'enabled': True,
        'pointer': None,
        'tree': None
    }
    profiler.set_meta(query='learn_process', api='http', environment=Config().get('environment'))
    with profiler.Context('learn_process'):
        from mindsdb.interfaces.database.database import DatabaseController

        try:
            predictor_record = db.Predictor.query.with_for_update().get(model_id)
            predictor_record.training_metadata['process_id'] = os.getpid()
            flag_modified(predictor_record, 'training_metadata')
            db.session.commit()

            target = problem_definition.get('target', None)
            training_data_df = None
            if data_integration_ref is not None:
                database_controller = DatabaseController()
                sql_session = make_sql_session()
                if data_integration_ref['type'] == 'integration':
                    integration_name = database_controller.get_integration(data_integration_ref['id'])['name']
                    query = Select(
                        targets=[Star()],
                        from_table=NativeQuery(
                            integration=Identifier(integration_name),
                            query=fetch_data_query
                        )
                    )
                    sqlquery = SQLQuery(query, session=sql_session)
                if data_integration_ref['type'] == 'system':
                    query = Select(
                        targets=[Star()],
                        from_table=NativeQuery(
                            integration=Identifier('log'),
                            query=fetch_data_query
                        )
                    )
                    sqlquery = SQLQuery(query, session=sql_session)
                elif data_integration_ref['type'] == 'view':
                    project = database_controller.get_project(project_name)
                    query_ast = parse_sql(fetch_data_query)
                    view_meta = project.query_view(query_ast)
                    sqlquery = SQLQuery(view_meta['query_ast'], session=sql_session)
                elif data_integration_ref['type'] == 'project':
                    query_ast = parse_sql(fetch_data_query)
                    sqlquery = SQLQuery(query_ast, session=sql_session)

                result = sqlquery.fetch(view='dataframe')
                training_data_df = result['result']

            training_data_columns_count, training_data_rows_count = 0, 0
            if training_data_df is not None:
                training_data_columns_count = len(training_data_df.columns)
                training_data_rows_count = len(training_data_df)

            predictor_record.training_data_columns_count = training_data_columns_count
            predictor_record.training_data_rows_count = training_data_rows_count
            db.session.commit()

            module = importlib.import_module(module_path)

            # check if module is imported successfully and raise exception if not
            if module.import_error is not None:
                raise module.import_error

            handlerStorage = HandlerStorage(integration_id)
            modelStorage = ModelStorage(model_id)
            modelStorage.fileStorage.push()     # FIXME

            kwargs = {}
            if base_model_id is not None:
                kwargs['base_model_storage'] = ModelStorage(base_model_id)
                kwargs['base_model_storage'].fileStorage.pull()
            ml_handler = module.Handler(
                engine_storage=handlerStorage,
                model_storage=modelStorage,
                **kwargs
            )
            handlers_cacher[predictor_record.id] = ml_handler

            if not ml_handler.generative:
                if training_data_df is not None and target not in training_data_df.columns:
                    raise Exception(
                        f'Prediction target "{target}" not found in training dataframe: {list(training_data_df.columns)}')

            # create new model
            if base_model_id is None:
                with profiler.Context('create'):
                    ml_handler.create(target, df=training_data_df, args=problem_definition)

            # fine-tune (partially train) existing model
            else:
                # load model from previous version, use it as starting point
                with profiler.Context('finetune'):
                    problem_definition['base_model_id'] = base_model_id
                    ml_handler.finetune(df=training_data_df, args=problem_definition)

            predictor_record.status = PREDICTOR_STATUS.COMPLETE
            predictor_record.active = set_active
            db.session.commit()
            # if retrain and set_active after success creation
            if set_active is True:
                models = get_model_records(
                    name=predictor_record.name,
                    project_id=predictor_record.project_id,
                    active=None
                )
                for model in models:
                    model.active = False
                models = [x for x in models if x.status == PREDICTOR_STATUS.COMPLETE]
                models.sort(key=lambda x: x.created_at)
                models[-1].active = True
        except Exception as e:
            logger.error(traceback.format_exc())
            error_message = format_exception_error(e)

            predictor_record = db.Predictor.query.with_for_update().get(model_id)
            predictor_record.data = {"error": error_message}
            predictor_record.status = PREDICTOR_STATUS.ERROR
            db.session.commit()

        predictor_record.training_stop_at = dt.datetime.now()
        db.session.commit()



"""content_storage

Revision ID: 3d5e70105df7
Revises: 87b2df2b83e1
Create Date: 2022-09-19 11:30:09.182435

"""
from pathlib import Path
import shutil

from mindsdb.utilities.config import Config


# revision identifiers, used by Alembic.
revision = '3d5e70105df7'
down_revision = '87b2df2b83e1'
branch_labels = None
depends_on = None


def upgrade():
    config = Config()
    is_cloud = config.get('cloud', False)
    if is_cloud is True:
        return

    storage_path = Path(config.paths['storage'])
    for item in storage_path.iterdir():
        if item.is_file() and item.name.startswith('predictor_'):
            original_name = item.name
            temp_file_path = item.parent / f'{original_name}_temp'
            item.replace(temp_file_path)
            item.mkdir(exist_ok=True)
            temp_file_path.replace(item / original_name)

    root_path = Path(config.paths['root'])
    if (root_path / 'datasources').is_dir():
        shutil.rmtree(root_path / 'datasources')
    if (root_path / 'integrations').is_dir():
        shutil.rmtree(root_path / 'integrations')
    if (root_path / 'predictors').is_dir():
        shutil.rmtree(root_path / 'predictors')


def downgrade():
    config = Config()
    is_cloud = config.get('cloud', False)
    if is_cloud is True:
        return

    storage_path = Path(config.paths['storage'])
    for item in storage_path.iterdir():
        if item.is_dir() and item.name.startswith('predictor_'):
            original_name = item.name
            if (item / original_name).is_file():
                temp_dir_path = item.parent / f'{original_name}_temp'
                item.replace(temp_dir_path)
                (temp_dir_path / original_name).replace(item.parent / original_name)
                shutil.rmtree(temp_dir_path)



from mindsdb.integrations.handlers.postgres_handler import Handler as PostgresHandler
from collections import OrderedDict
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


class TimeScaleDBHandler(PostgresHandler):
    name = 'timescaledb'

    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)




connection_args = OrderedDict(
    host={
        'type': ARG_TYPE.STR,
        'description': 'The host name or IP address of the TimeScaleDB server/database.'
    },
    database={
        'type': ARG_TYPE.STR,
        'description': """
            The database name to use when connecting with the TimeScaleDB server.
        """
    },
    user={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with the TimeScaleDB server.'
    },
    password={
        'type': ARG_TYPE.STR,
        'description': 'The password to authenticate the user with the TimeScaleDB server.'
    },
    port={
        'type': ARG_TYPE.INT,
        'description': 'Specify port to connect TimeScaleDB '
    }
)

connection_args_example = OrderedDict(
    host='127.0.0.1',
    port=36806,
    password='P455W0rD',
    user='tsdbadmin',
    database="tsdb"
)



from typing import Type
import re
from mindsdb_sql_parser import parse_sql
from pydantic import BaseModel, Field
from langchain_core.tools import BaseTool


class _MindsDBSQLParserToolInput(BaseModel):
    tool_input: str = Field("", description="A SQL query to validate.")


class MindsDBSQLParserTool(BaseTool):
    name: str = "mindsdb_sql_parser_tool"
    description: str = "Parse a SQL query to check it is valid MindsDB SQL."
    args_schema: Type[BaseModel] = _MindsDBSQLParserToolInput

    def _clean_query(self, query: str) -> str:
        # Sometimes LLM can input markdown into query tools.
        cmd = re.sub(r'```(sql)?', '', query)
        return cmd

    def _query_options(self, query):
        yield query
        if '\\_' in query:
            yield query.replace('\\_', '_')

    def _run(self, query: str):
        """Validate the SQL query."""
        clean_query = self._clean_query(query)
        for query in self._query_options(clean_query):
            try:
                ast_query = parse_sql(query)
                return "".join(f"valid query: {ast_query.to_string()}")
            except Exception as e:
                error = "".join(f"invalid query, with error: {e}")
                continue
        return error



import os
import pytest
import pandas as pd
from unittest.mock import patch

from tests.unit.ml_handlers.base_ml_test import BaseMLAPITest


@pytest.mark.skipif(os.environ.get('MDB_TEST_MINDS_ENDPOINT_API_KEY') is None, reason='Missing API key!')
class TestMindsEndpoint(BaseMLAPITest):
    """
    Integration tests for Minds Endpoint engine.
    """

    # TODO: Should random names be generated for the project, model etc.?
    # TODO: Are the resources created being cleaned up after the test?

    def setup_method(self):
        """
        Setup test environment by creating a project and a Minds Endpoint engine.
        """

        super().setup_method()
        self.run_sql("CREATE DATABASE proj")
        self.run_sql(
            f"""
            CREATE ML_ENGINE minds_endpoint_engine
            FROM minds_endpoint
            USING
            minds_endpoint_api_key = '{self.get_api_key('MDB_TEST_MINDS_ENDPOINT_API_KEY')}';
            """
        )

    def test_create_model_raises_exception_with_invalid_model_parameter(self):
        """
        Test for invalid parameter during model creation.
        """

        self.run_sql(
            f"""
            CREATE MODEL proj.test_minds_endpoint_invalid_parameter_model
            PREDICT answer
            USING
                engine='minds_endpoint_engine',
                model_name='this-model-does-not-exist',
                prompt_template='dummy_prompt_template',
                minds_endpoint_api_key='{self.get_api_key('MDB_TEST_MINDS_ENDPOINT_API_KEY')}';
            """
        )
        with pytest.raises(Exception):
            self.wait_predictor("proj", "test_minds_endpoint_invalid_model")

    def test_create_model_raises_exception_with_unknown_model_argument(self):
        """
        Test for unknown argument during model creation.
        """

        self.run_sql(
            f"""
            CREATE MODEL proj.test_minds_endpoint_unknown_argument_model
            PREDICT answer
            USING
                engine='minds_endpoint_engine',
                prompt_template='dummy_prompt_template',
                minds_endpoint_api_key='{self.get_api_key('MDB_TEST_MINDS_ENDPOINT_API_KEY')}',
                evidently_wrong_argument='wrong value';
            """
        )
        with pytest.raises(Exception):
            self.wait_predictor("proj", "test_minds_endpoint_unknown_argument_model")

    def test_select_runs_no_errors_on_chat_completion_question_answering_single(self):
        """
        Test for a valid answer to a question answering task (chat completion).
        """

        self.run_sql(
            f"""
            CREATE MODEL proj.test_minds_endpoint_single_qa
            PREDICT answer
            USING
                engine='minds_endpoint_engine',
                question_column='question',
                minds_endpoint_api_key='{self.get_api_key('MDB_TEST_MINDS_ENDPOINT_API_KEY')}';
            """
        )
        self.wait_predictor("proj", "test_minds_endpoint_single_qa")

        result_df = self.run_sql(
            """
            SELECT answer
            FROM proj.test_minds_endpoint_single_qa
            WHERE question = 'What is the capital of Sweden?';
            """
        )

        assert "stockholm" in result_df["answer"].iloc[0].lower()

    @patch("mindsdb.integrations.handlers.postgres_handler.Handler")
    def test_select_runs_no_errors_on_chat_completion_question_answering_bulk(self, mock_postgres_handler):
        """
        Test for valid answers to bulk questions in a question answering task (chat completion).
        """

        df = pd.DataFrame.from_dict({"question": [
            "What is the capital of Sweden?",
            "What is the second planet in the solar system?"
        ]})
        self.set_handler(mock_postgres_handler, name="pg", tables={"df": df})

        self.run_sql(
            f"""
            CREATE MODEL proj.test_minds_endpoint_bulk_qa
            PREDICT answer
            USING
                engine='minds_endpoint_engine',
                question_column='question',
                minds_endpoint_api_key='{self.get_api_key('MDB_TEST_MINDS_ENDPOINT_API_KEY')}';
            """
        )
        self.wait_predictor("proj", "test_minds_endpoint_bulk_qa")

        result_df = self.run_sql(
            """
            SELECT p.answer
            FROM pg.df as t
            JOIN proj.test_minds_endpoint_bulk_qa as p;
        """
        )

        assert "stockholm" in result_df["answer"].iloc[0].lower()
        assert "venus" in result_df["answer"].iloc[1].lower()

    def test_select_runs_no_errors_on_embeddings_completion_single(self):
        """
        Test for a valid answer to a question answering task (chat completion).
        """

        self.run_sql(
            f"""
            CREATE MODEL proj.test_minds_endpoint_single_embeddings
            PREDICT embeddings
            USING
                engine='minds_endpoint_engine',
                question_column='text',
                mode='embedding',
                minds_endpoint_api_key='{self.get_api_key('MDB_TEST_MINDS_ENDPOINT_API_KEY')}';
            """
        )
        self.wait_predictor("proj", "test_minds_endpoint_single_embeddings")

        result_df = self.run_sql(
            """
            SELECT embeddings
            FROM proj.test_minds_endpoint_single_embeddings
            WHERE text = 'MindsDB';
            """
        )

        assert isinstance(result_df["embeddings"].iloc[0], list)



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    host={
        'type': ARG_TYPE.STR,
        'description': "The host name or IP address of the Apache Ignite cluster's node.",
        'required': True,
        'label': 'Host'
    },
    port={
        'type': ARG_TYPE.INT,
        'description': "The TCP/IP port of the Apache Ignite cluster's node. Must be an integer.",
        'required': True,
        'label': 'Port'
    },
    username={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with the Apache Ignite cluster. This parameter is optional. Default: None.',
        'required': True,
        'label': 'User'
    },
    password={
        'type': ARG_TYPE.PWD,
        'description': 'The password to authenticate the user with the Apache Ignite cluster. This parameter is optional. Default: None.',
        'required': True,
        'label': 'Password',
        'secret': True
    },
    schema={
        'type': ARG_TYPE.STR,
        'description': 'Schema to use for the connection to the Apache Ignite cluster. This parameter is optional. Default: PUBLIC.',
        'required': True,
        'label': 'Schema'
    }
)

connection_args_example = OrderedDict(
    host='127.0.0.1',
    port='10800',
    username='root',
    password='password',
    schema='schema'
)



from typing import List

from langchain_community.agent_toolkits.sql.toolkit import SQLDatabaseToolkit
from langchain_community.tools import ListSQLDatabaseTool, InfoSQLDatabaseTool, QuerySQLDataBaseTool
from langchain_core.tools import BaseTool

from mindsdb.interfaces.skills.custom.text2sql.mindsdb_sql_tool import MindsDBSQLParserTool


class MindsDBSQLToolkit(SQLDatabaseToolkit):

    def get_tools(self, prefix='') -> List[BaseTool]:
        """Get the tools in the toolkit."""
        list_sql_database_tool = ListSQLDatabaseTool(name=f'sql_db_list_tables{prefix}', db=self.db)

        info_sql_database_tool_description = (
            "Input: A comma-separated list of tables. Output: Schema and sample rows for those tables. "
            f"Ensure tables exist by calling {list_sql_database_tool.name} first. "
            "Use this tool to investigate table schemas for needed columns. "
            "Get sample data with 'SELECT * FROM table LIMIT 3' before answering questions. "
            "Example Input: table1, table2, table3"
        )
        info_sql_database_tool = InfoSQLDatabaseTool(
            name=f'sql_db_schema{prefix}',
            db=self.db, description=info_sql_database_tool_description
        )

        query_sql_database_tool_description = (
            "Input: A detailed SQL query. Output: Database result or error message. "
            "For errors, rewrite and retry the query. For 'Unknown column' errors, use "
            f"{info_sql_database_tool.name} to check table fields. "
            "This system is a highly intelligent and reliable PostgreSQL SQL skill designed to work with databases. "
            "Follow these instructions with utmost precision: "
            "1. Query Output Format: "
            "   - Always return results in well-formatted **Markdown tables**. "
            "   - Ensure clarity and proper structure for easy readability. "
            "2. Sample Data: "
            "   - Before answering a question, if you don't have sample data about a table, **always** get sample data using `SELECT * FROM table LIMIT 3` from the tables you believe are relevant to formulating your answers. "
            "3. Categorical Data: "
            "   - Whenever working with a column where values seem categorical, especially when filtering with `WHERE col = 'value'`, `WHERE col IN (list of values)`, or `WHERE col NOT IN (list of values)`, **always** retrieve the distinct values first. "
            "   - Before writing your main query, always run `SELECT DISTINCT col` to fetch a list of unique values from that column. This step is mandatory to ensure accurate queries and responses. "
            "4. Result Limiting and Counting: "
            "   - Unless instructed otherwise by the user, always run a count on the final query first using `SELECT COUNT(*)`. "
            "   - If the count is greater than 10, limit the query to return only 10 results initially. "
            "   - **Always** inform the user of the total number of results available and specify that you are providing the first 10 results. "
            "   - Let the user know they can request additional results and/or specify how they would like the results ordered or grouped. "
            "5. Date Handling: "
            "   - **Always** use PostgreSQL-compatible `CURRENT_DATE` or `NOW()` functions when working with dates—never assume or guess the current date. "
            "   - For any date-related comparisons in the query, *always* ensure that your query casts the column being compared using `column_name::DATE [operator] ..` "
            "   - Do not compare date values without casting columns to date. "
            "   - For date interval operations, use Interval units as keywords. You can use keywords to specify units like days, hours, months, years, etc., directly without quotes. Examples: "
            "     SELECT NOW() + INTERVAL 5 DAY; "
            "     SELECT NOW() - INTERVAL 3 HOUR; "
            "     SELECT NOW() + INTERVAL 2 MONTH + INTERVAL 3 DAY; "
            "     SELECT NOW() - INTERVAL 1 YEAR; "
            "6. Query Best Practices: "
            "   - Query only necessary columns, not all. "
            "   - Use only existing column names from correct tables. "
            "   - Use database-specific syntax for date operations. "
            "7. Error Handling: "
            "   - For errors, rewrite and retry the query. "
            "   - For 'Unknown column' errors, check table fields using info_sql_database_tool. "
            "Adhere to these guidelines for all queries and responses. Ask for clarification if needed."
        )

        query_sql_database_tool = QuerySQLDataBaseTool(
            name=f'sql_db_query{prefix}',
            db=self.db, description=query_sql_database_tool_description
        )

        mindsdb_sql_parser_tool_description = (
            "Use this tool to ensure a SQL query passes the MindsDB SQL parser. "
            "If the query is not correct, it will be corrected and returned. Use the new query. "
            "If the query can't be corrected, an error is returned. In this case, rewrite and retry. "
            "If the query is correct, it will be parsed and returned. "
            f"ALWAYS run this tool before executing a query with {query_sql_database_tool.name}. "
        )
        mindsdb_sql_parser_tool = MindsDBSQLParserTool(
            name=f'mindsdb_sql_parser_tool{prefix}',
            description=mindsdb_sql_parser_tool_description
        )

        return [
            query_sql_database_tool,
            info_sql_database_tool,
            list_sql_database_tool,
            mindsdb_sql_parser_tool,
        ]



from enum import Enum
from typing import List, Union, Any, Optional, Dict

from langchain_community.vectorstores.chroma import Chroma
from langchain_community.vectorstores.pgvector import PGVector
from langchain_core.documents import Document
from langchain_core.embeddings import Embeddings
from langchain_core.language_models import BaseChatModel
from langchain_core.vectorstores import VectorStore
from langchain_core.stores import BaseStore
from pydantic import BaseModel, Field, field_validator
from langchain_text_splitters import TextSplitter

DEFAULT_COLLECTION_NAME = 'default_collection'

# Multi retriever specific
DEFAULT_ID_KEY = "doc_id"
DEFAULT_MAX_CONCURRENCY = 5
DEFAULT_K = 20

DEFAULT_CARDINALITY_THRESHOLD = 40
DEFAULT_MAX_SUMMARIZATION_TOKENS = 4000
DEFAULT_CHUNK_SIZE = 1000
DEFAULT_CHUNK_OVERLAP = 200
DEFAULT_POOL_RECYCLE = 3600
DEFAULT_LLM_MODEL = "gpt-4o"
DEFAULT_LLM_MODEL_PROVIDER = "openai"
DEFAULT_CONTENT_COLUMN_NAME = "body"
DEFAULT_DATASET_DESCRIPTION = "email inbox"
DEFAULT_TEST_TABLE_NAME = "test_email"
DEFAULT_VECTOR_STORE = Chroma
DEFAULT_RERANKER_FLAG = False
DEFAULT_RERANKING_MODEL = "gpt-4o"
DEFAULT_LLM_ENDPOINT = "https://api.openai.com/v1"
DEFAULT_AUTO_META_PROMPT_TEMPLATE = """
Below is a json representation of a table with information about {description}.
Return a JSON list with an entry for each column. Each entry should have
{{"name": "column name", "description": "column description", "type": "column data type"}}
\n\n{dataframe}\n\nJSON:\n
"""
DEFAULT_RAG_PROMPT_TEMPLATE = '''You are an assistant for
question-answering tasks. Use the following pieces of retrieved context
to answer the question. If you don't know the answer, just say that you
don't know. Use two sentences maximum and keep the answer concise.
Question: {question}
Context: {context}
Answer:'''

DEFAULT_QA_GENERATION_PROMPT_TEMPLATE = '''You are an assistant for
generating sample questions and answers from the given document and metadata. Given
a document and its metadata as context, generate a question and answer from that document and its metadata.

The document will be a string. The metadata will be a JSON string. You need
to parse the JSON to understand it.

Generate a question that requires BOTH the document and metadata to answer, if possible.
Otherwise, generate a question that requires ONLY the document to answer.

Return a JSON dictionary with the question and answer like this:
{{ "question": <the full generated question>, "answer": <the full generated answer> }}

Make sure the JSON string is valid before returning it. You must return the question and answer
in the specified JSON format no matter what.

Document: {document}
Metadata: {metadata}
Answer:'''

DEFAULT_MAP_PROMPT_TEMPLATE = '''The following is a set of documents
{docs}
Based on this list of docs, please summarize based on the user input.

User input: {input}

Helpful Answer:'''

DEFAULT_REDUCE_PROMPT_TEMPLATE = '''The following is set of summaries:
{docs}
Take these and distill it into a final, consolidated summary related to the user input.

User input: {input}

Helpful Answer:'''


class LLMConfig(BaseModel):
    model_name: str = Field(default=DEFAULT_LLM_MODEL, description='LLM model to use for generation')
    provider: str = Field(default=DEFAULT_LLM_MODEL_PROVIDER, description='LLM model provider to use for generation')
    params: Dict[str, Any] = {}


class MultiVectorRetrieverMode(Enum):
    """
    Enum for MultiVectorRetriever types.
    """
    SPLIT = "split"
    SUMMARIZE = "summarize"
    BOTH = "both"


class VectorStoreType(Enum):
    CHROMA = 'chromadb'
    PGVECTOR = 'pgvector'


vector_store_map = {
    VectorStoreType.CHROMA: Chroma,
    VectorStoreType.PGVECTOR: PGVector
}


class VectorStoreConfig(BaseModel):
    vector_store_type: VectorStoreType = VectorStoreType.CHROMA
    persist_directory: str = None
    collection_name: str = DEFAULT_COLLECTION_NAME
    connection_string: str = None
    kb_table: Any = None

    class Config:
        arbitrary_types_allowed = True
        extra = "forbid"


class RetrieverType(Enum):
    VECTOR_STORE = 'vector_store'
    AUTO = 'auto'
    MULTI = 'multi'


class SearchType(Enum):
    """
    Enum for vector store search types.
    """
    SIMILARITY = "similarity"
    MMR = "mmr"
    SIMILARITY_SCORE_THRESHOLD = "similarity_score_threshold"


class SearchKwargs(BaseModel):
    k: int = Field(
        default=DEFAULT_K,
        description="Amount of documents to return",
        ge=1
    )
    filter: Optional[Dict[str, Any]] = Field(
        default=None,
        description="Filter by document metadata"
    )
    # For similarity_score_threshold search type
    score_threshold: Optional[float] = Field(
        default=None,
        description="Minimum relevance threshold for similarity_score_threshold search",
        ge=0.0,
        le=1.0
    )
    # For MMR search type
    fetch_k: Optional[int] = Field(
        default=None,
        description="Amount of documents to pass to MMR algorithm",
        ge=1
    )
    lambda_mult: Optional[float] = Field(
        default=None,
        description="Diversity of results returned by MMR (1=min diversity, 0=max)",
        ge=0.0,
        le=1.0
    )

    def model_dump(self, *args, **kwargs):
        # Override model_dump to exclude None values by default
        kwargs['exclude_none'] = True
        return super().model_dump(*args, **kwargs)


class SummarizationConfig(BaseModel):
    llm_config: LLMConfig = Field(
        default_factory=LLMConfig,
        description="LLM configuration to use for summarization"
    )
    map_prompt_template: str = Field(
        default=DEFAULT_MAP_PROMPT_TEMPLATE,
        description="Prompt for an LLM to summarize a single document"
    )
    reduce_prompt_template: str = Field(
        default=DEFAULT_REDUCE_PROMPT_TEMPLATE,
        description="Prompt for an LLM to summarize a set of summaries of documents into one"
    )
    max_summarization_tokens: int = Field(
        default=DEFAULT_MAX_SUMMARIZATION_TOKENS,
        description="Max number of tokens for summarized documents"
    )


class RerankerConfig(BaseModel):
    model: str = DEFAULT_RERANKING_MODEL
    base_url: str = DEFAULT_LLM_ENDPOINT
    filtering_threshold: float = 0.99
    num_docs_to_keep: Optional[int] = None


class RAGPipelineModel(BaseModel):
    documents: Optional[List[Document]] = Field(
        default=None,
        description="List of documents"
    )

    vector_store_config: VectorStoreConfig = Field(
        default_factory=VectorStoreConfig,
        description="Vector store configuration"
    )

    llm: Optional[BaseChatModel] = Field(
        default=None,
        description="Language model"
    )
    llm_model_name: str = Field(
        default=DEFAULT_LLM_MODEL,
        description="Language model name"
    )
    llm_provider: Optional[str] = Field(
        default=None,
        description="Language model provider"
    )

    vector_store: VectorStore = Field(
        default_factory=lambda: vector_store_map[VectorStoreConfig().vector_store_type],
        description="Vector store"
    )
    db_connection_string: Optional[str] = Field(
        default=None,
        description="Database connection string"
    )
    table_name: str = Field(
        default=DEFAULT_TEST_TABLE_NAME,
        description="Table name"
    )
    embedding_model: Optional[Embeddings] = Field(
        default=None,
        description="Embedding model"
    )
    rag_prompt_template: str = Field(
        default=DEFAULT_RAG_PROMPT_TEMPLATE,
        description="RAG prompt template"
    )
    retriever_prompt_template: Optional[Union[str, dict]] = Field(
        default=None,
        description="Retriever prompt template"
    )
    retriever_type: RetrieverType = Field(
        default=RetrieverType.VECTOR_STORE,
        description="Retriever type"
    )
    search_type: SearchType = Field(
        default=SearchType.SIMILARITY,
        description="Type of search to perform"
    )
    search_kwargs: SearchKwargs = Field(
        default_factory=SearchKwargs,
        description="Search configuration for the retriever"
    )
    summarization_config: Optional[SummarizationConfig] = Field(
        default=None,
        description="Configuration for summarizing retrieved documents as context"
    )

    # Multi retriever specific
    multi_retriever_mode: MultiVectorRetrieverMode = Field(
        default=MultiVectorRetrieverMode.BOTH,
        description="Multi retriever mode"
    )
    max_concurrency: int = Field(
        default=DEFAULT_MAX_CONCURRENCY,
        description="Maximum concurrency"
    )
    id_key: int = Field(
        default=DEFAULT_ID_KEY,
        description="ID key"
    )
    parent_store: Optional[BaseStore] = Field(
        default=None,
        description="Parent store"
    )
    text_splitter: Optional[TextSplitter] = Field(
        default=None,
        description="Text splitter"
    )
    chunk_size: int = Field(
        default=DEFAULT_CHUNK_SIZE,
        description="Chunk size"
    )
    chunk_overlap: int = Field(
        default=DEFAULT_CHUNK_OVERLAP,
        description="Chunk overlap"
    )

    # Auto retriever specific
    auto_retriever_filter_columns: Optional[List[str]] = Field(
        default=None,
        description="Filter columns"
    )
    cardinality_threshold: int = Field(
        default=DEFAULT_CARDINALITY_THRESHOLD,
        description="Cardinality threshold"
    )
    content_column_name: str = Field(
        default=DEFAULT_CONTENT_COLUMN_NAME,
        description="Content column name (the column we will get embeddings)"
    )
    dataset_description: str = Field(
        default=DEFAULT_DATASET_DESCRIPTION,
        description="Description of the dataset"
    )
    reranker: bool = Field(
        default=DEFAULT_RERANKER_FLAG,
        description="Whether to use reranker"
    )
    reranker_config: RerankerConfig = Field(
        default_factory=RerankerConfig,
        description="Reranker configuration"
    )

    class Config:
        arbitrary_types_allowed = True
        extra = "forbid"

        json_schema_extra = {
            "example": {
                "retriever_type": RetrieverType.VECTOR_STORE.value,
                "multi_retriever_mode": MultiVectorRetrieverMode.BOTH.value,
                # add more examples here
            }
        }

    @classmethod
    def get_field_names(cls):
        return list(cls.model_fields.keys())

    @field_validator('search_kwargs')
    @classmethod
    def validate_search_kwargs(cls, v: SearchKwargs, info) -> SearchKwargs:
        search_type = info.data.get('search_type', SearchType.SIMILARITY)

        # Validate MMR-specific parameters
        if search_type == SearchType.MMR:
            if v.fetch_k is not None and v.fetch_k <= v.k:
                raise ValueError("fetch_k must be greater than k")
            if v.lambda_mult is not None and (v.lambda_mult < 0 or v.lambda_mult > 1):
                raise ValueError("lambda_mult must be between 0 and 1")
            if v.fetch_k is None and v.lambda_mult is not None:
                raise ValueError("fetch_k is required when using lambda_mult with MMR search type")
            if v.lambda_mult is None and v.fetch_k is not None:
                raise ValueError("lambda_mult is required when using fetch_k with MMR search type")
        elif search_type != SearchType.MMR:
            if v.fetch_k is not None:
                raise ValueError("fetch_k is only valid for MMR search type")
            if v.lambda_mult is not None:
                raise ValueError("lambda_mult is only valid for MMR search type")

        # Validate similarity_score_threshold parameters
        if search_type == SearchType.SIMILARITY_SCORE_THRESHOLD:
            if v.score_threshold is not None and (v.score_threshold < 0 or v.score_threshold > 1):
                raise ValueError("score_threshold must be between 0 and 1")
            if v.score_threshold is None:
                raise ValueError("score_threshold is required for similarity_score_threshold search type")
        elif search_type != SearchType.SIMILARITY_SCORE_THRESHOLD and v.score_threshold is not None:
            raise ValueError("score_threshold is only valid for similarity_score_threshold search type")

        return v
